#!/usr/bin/env python
# -*- coding: utf-8 -*-
# kate: word-wrap off; remove-trailing-spaces all;

"""
atropos version {version}

Atropos removes adapter sequences from high-throughput sequencing reads.

Usage:
    atropos -a ADAPTER [options] [-o output.fastq] input.fastq

For paired-end reads:
    atropos -a ADAPT1 -A ADAPT2 [options] -o out1.fastq -p out2.fastq in1.fastq in2.fastq

Replace "ADAPTER" with the actual sequence of your 3' adapter. IUPAC wildcard
characters are supported. The reverse complement is *not* automatically
searched. All reads from input.fastq will be written to output.fastq with the
adapter sequence removed. Adapter matching is error-tolerant. Multiple adapter
sequences can be given (use further -a options), but only the best-matching
adapter will be removed.

Input may also be in FASTA format. Compressed input and output is supported and
auto-detected from the file name (.gz, .xz, .bz2). Use the file name '-' for
standard input/output. Without the -o option, output is sent to standard output.

Use "atropos --help" to see all command-line options.
See http://atropos.readthedocs.org/ for full documentation.

Cutadapt (https://github.com/marcelm/cutadapt) was developed by Marcel Martin,
"Cutadapt Removes Adapter Sequences From High-Throughput Sequencing Reads,"
EMBnet Journal, 2011, 17(1):10-12.

Atropos is a fork of Cutadapt 1.10 (
https://github.com/marcelm/cutadapt/tree/2f3cc0717aa9ff1e0326ea6bcb36b712950d4999)
by John Didion, "Atropos: sensitive, specific, and speedy trimming of NGS reads,"
in prep.
"""

# Print a helpful error message if the extension modules cannot be imported.
from atropos import *
check_importability()

from atropos import __version__
from atropos.seqio import (open_reader, UnknownFileType, Formatters, RestFormatter,
                            InfoFormatter, WildcardFormatter, FormatError, Writers)
from atropos.adapters import AdapterParser
from atropos.modifiers import *
from atropos.filters import *
from atropos.report import *

import argparse
from collections import defaultdict
import copy
import logging
import os
import platform
import sys
import time

def main(cmdlineargs=None, default_outfile="-"):
    """
    Main function that evaluates command-line parameters and iterates over all reads.
    
    :param cmdlineargs: iterable of command line arguments; `None` is equivalent to
    `sys.argv[1:]`
    :param default_outfile: the file to which trimmed reads are sent if the ``-o``
    parameter is not used.
    """
    orig_args = copy.copy(cmdlineargs or sys.argv)
    parser = get_argument_parser()
    options = parser.parse_args(args=cmdlineargs)
    
    # Setup logging only if there are not already any handlers (can happen when
    # this function is being called externally such as from unit tests)
    if not logging.root.handlers:
        level = options.log_level or ("ERROR" if options.quiet else "INFO")
        setup_logging(stdout=bool(options.output), level=level)
    
    paired = validate_options(options, parser)
    reader, qualities, has_qual_file = create_reader(options.inputs, options, paired, parser)
    modifiers, num_adapters = create_modifiers(options, paired, qualities, has_qual_file, parser)
    min_affected = 2 if options.pair_filter == 'both' else 1
    filters = create_filters(options, paired, min_affected)
    formatters, force_create = create_formatters(options, qualities, default_outfile)
    writers = Writers(force_create)
    
    logger = logging.getLogger()
    logger.info("This is atropos %s with Python %s", __version__, platform.python_version())
    logger.info("Command line parameters: %s", " ".join(orig_args))
    logger.info("Trimming %s adapter%s with at most %.1f%% errors in %s mode ...",
        num_adapters, 's' if num_adapters > 1 else '', options.error_rate * 100,
        { False: 'single-end', 'first': 'paired-end legacy', 'both': 'paired-end' }[paired])
    if paired == 'first' and (len(modifiers.modifiers2) > 0 or options.quality_cutoff):
        import textwrap
        logger.warning('\n'.join(textwrap.wrap('WARNING: Requested read '
            'modifications are applied only to the first '
            'read since backwards compatibility mode is enabled. '
            'To modify both reads, also use any of the -A/-B/-G/-U options. '
            'Use a dummy adapter sequence when necessary: -A XXX')))
    
    start_wallclock_time = time.time()
    start_cpu_time = time.clock()
    
    if options.threads is None:
        # Run single-threaded version
        rc, summary = run_serial(reader, modifiers, filters, formatters, writers)
    else:
        # Run multiprocessing version
        import atropos.multicore
        rc, summary = atropos.multicore.run_parallel(
            reader, modifiers, filters, formatters, writers, options.threads,
            options.process_timeout, options.preserve_order, options.read_queue_size,
            options.result_queue_size, not options.no_writer_process, options.worker_compression)
    
    if rc != 0:
        sys.exit(rc)
    
    stop_wallclock_time = time.time()
    stop_cpu_time = time.clock()
    report = print_report(
        paired,
        options,
        stop_wallclock_time - start_wallclock_time,
        stop_cpu_time - start_cpu_time,
        summary,
        modifiers.get_trimmer_classes())
    
def get_argument_parser():
    parser = argparse.ArgumentParser(usage=__doc__.lstrip().format(version=__version__))
    parser.add_argument("inputs", nargs=argparse.REMAINDER)
    
    parser.add_argument("--debug", action='store_true', default=False,
        help="Print debugging information. (no)")
    parser.add_argument("-f", "--format",
        help="Input file format; can be either 'fasta', 'fastq' or 'sra-fastq'. "
            "Ignored when reading csfasta/qual files. (auto-detect "
            "from file name extension)")
    parser.add_argument("--max-reads", default=None,
        help="Maximum number of reads/pairs to process (no max)")
    parser.add_argument("--progress", default=None,
        help="Show progress. bar = show progress bar; msg = show a status "
             "message. (no)")
    
    group = parser.add_argument_group("Finding adapters:",
        description="Parameters -a, -g, -b specify adapters to be removed from "
            "each read (or from the first read in a pair if data is paired). "
            "If specified multiple times, only the best matching adapter is "
            "trimmed (but see the --times option). When the special notation "
            "'file:FILE' is used, adapter sequences are read from the given "
            "FASTA file.")
    group.add_argument("-a", "--adapter", action="append", default=[], metavar="ADAPTER",
        dest="adapters",
        help="Sequence of an adapter ligated to the 3' end (paired data: of the "
            "first read). The adapter and subsequent bases are trimmed. If a "
            "'$' character is appended ('anchoring'), the adapter is only "
            "found if it is a suffix of the read. (none)")
    group.add_argument("-g", "--front", action="append", default=[], metavar="ADAPTER",
        help="Sequence of an adapter ligated to the 5' end (paired data: of the "
            "first read). The adapter and any preceding bases are trimmed. "
            "Partial matches at the 5' end are allowed. If a '^' character is "
            "prepended ('anchoring'), the adapter is only found if it is a "
            "prefix of the read. (none)")
    group.add_argument("-b", "--anywhere", action="append", default=[], metavar="ADAPTER",
        help="Sequence of an adapter that may be ligated to the 5' or 3' end "
            "(paired data: of the first read). Both types of matches as "
            "described under -a und -g are allowed. If the first base of the "
            "read is part of the match, the behavior is as with -g, otherwise "
            "as with -a. This option is mostly for rescuing failed library "
            "preparations - do not use if you know which end your adapter was "
            "ligated to! (none)")
    group.add_argument("-e", "--error-rate", type=float, default=None,
        help="Maximum allowed error rate (no. of errors divided by the length "
            "of the matching region). (0.1)")
    group.add_argument("--no-indels", action='store_false', dest='indels', default=True,
        help="Allow only mismatches in alignments. "
            "(allow both mismatches and indels)")
    group.add_argument("-n", "--times", type=int, metavar="COUNT", default=1,
        help="Remove up to COUNT adapters from each read. (1)")
    group.add_argument("-O", "--overlap", type=int, metavar="MINLENGTH", default=3,
        help="If the overlap between the read and the adapter is shorter than "
            "MINLENGTH, the read is not modified. Reduces the no. of bases "
            "trimmed due to random adapter matches. (3)")
    group.add_argument("--match-read-wildcards", action="store_true", default=False,
        help="Interpret IUPAC wildcards in reads. (no)")
    group.add_argument("-N", "--no-match-adapter-wildcards", action="store_false",
        default=True, dest='match_adapter_wildcards',
        help="Do not interpret IUPAC wildcards in adapters. (no)")
    group.add_argument("--no-trim", dest='action', action='store_const', const=None,
        help="Match and redirect reads to output/untrimmed-output as usual, "
            "but do not remove adapters. (no)")
    group.add_argument("--mask-adapter", dest='action', action='store_const', const='mask',
        help="Mask adapters with 'N' characters instead of trimming them. (no)")
    group.add_argument("-R", "--merge-overlapping", action="store_true", default=False,
        help="Merge read pairs that overlap into a single sequence. This is an experimental "
             "and currently only works with read pairs where an adapter match was found at "
             "the 3' ends of both reads. The minimum required overlap is controled by "
             "--merge-min-overlap, and the mimimum error rate in the alignment is controlled "
             "by --error-rate. (no)")
    group.add_argument("--merge-min-overlap", type=float, default=0.9,
        help="The minimum overlap between reads required for merging. If this number is (0,1.0], "
             "it specifies the minimum length as the fraction of the length of the *shorter* read "
             "in the pair; otherwise it specifies the minimum number of overlapping base pairs ("
             "with an absolute minimum of 2 bp). (0.9)")
    #group.add_argument("--merge-mismatches", choices=("best", "N", "random"), default="best",
    #	help="How to handle mismatches while merging; best: select the base with the best quality; "
    #		 "N: set the base to N; random: select one of the two bases at random. Note that if "
    #		 "exactly one base is ambiguous, the non-ambiguous base is always used.")
    
    group = parser.add_argument_group("Additional read modifications")
    group.add_argument("-u", "--cut", action='append', default=[], type=int, metavar="LENGTH",
        help="Remove bases from each read (first read only if paired). "
            "If LENGTH is positive, remove bases from the beginning. "
            "If LENGTH is negative, remove bases from the end. "
            "Can be used twice if LENGTHs have different signs. (no)")
    group.add_argument("-q", "--quality-cutoff", default=None, metavar="[5'CUTOFF,]3'CUTOFF",
        help="Trim low-quality bases from 5' and/or 3' ends of each read before "
            "adapter removal. Applied to both reads if data is paired. If one "
            "value is given, only the 3' end is trimmed. If two "
            "comma-separated cutoffs are given, the 5' end is trimmed with "
            "the first cutoff, the 3' end with the second. (no)")
    group.add_argument("-i", "--cut-min", action='append', default=[], type=int, metavar="LENGTH",
        help="Similar to -u, except that cutting is done AFTER adapter trimming, and only "
            "if a minimum of LENGTH bases was not already removed. (no)")
    group.add_argument("--nextseq-trim", type=int, default=None, metavar="3'CUTOFF",
        help="NextSeq-specific quality trimming (each read). Trims also dark "
            "cycles appearing as high-quality G bases (EXPERIMENTAL). (no)")
    group.add_argument("--quality-base", type=int, default=33,
        help="Assume that quality values in FASTQ are encoded as ascii(quality "
            "+ QUALITY_BASE). This needs to be set to 64 for some old Illumina "
            "FASTQ files. (33)")
    group.add_argument("--trim-n", action='store_true', default=False,
        help="Trim N's on ends of reads. (no)")
    group.add_argument("-x", "--prefix", default='',
        help="Add this prefix to read names. Use {name} to insert the name of "
             "the matching adapter. (no)")
    group.add_argument("-y", "--suffix", default='',
        help="Add this suffix to read names; can also include {name}. (no)")
    group.add_argument("--strip-suffix", action='append', default=[],
        help="Remove this suffix from read names if present. Can be given "
             "multiple times. (no)")
    group.add_argument("--length-tag", metavar="TAG",
        help="Search for TAG followed by a decimal number in the description "
            "field of the read. Replace the decimal number with the correct "
            "length of the trimmed read. For example, use --length-tag 'length=' "
            "to correct fields like 'length=123'. (no)")
    
    group = parser.add_argument_group("Filtering of processed reads")
    group.add_argument("--discard-trimmed", "--discard", action='store_true', default=False,
        help="Discard reads that contain an adapter. Also use -O to avoid "
            "discarding too many randomly matching reads! (no)")
    group.add_argument("--discard-untrimmed", "--trimmed-only", action='store_true', default=False,
        help="Discard reads that do not contain the adapter. (no)")
    group.add_argument("-m", "--minimum-length", type=int, default=None, metavar="LENGTH",
        help="Discard trimmed reads that are shorter than LENGTH. Reads that "
            "are too short even before adapter removal are also discarded. In "
            "colorspace, an initial primer is not counted. (0)")
    group.add_argument("-M", "--maximum-length", type=int, default=sys.maxsize, metavar="LENGTH",
        help="Discard trimmed reads that are longer than LENGTH. "
            "Reads that are too long even before adapter removal "
            "are also discarded. In colorspace, an initial primer "
            "is not counted. (no limit)")
    group.add_argument("--max-n", type=float, default=-1.0, metavar="COUNT",
        help="Discard reads with too many N bases. If COUNT is an integer, it "
            "is treated as the absolute number of N bases. If it is between 0 "
            "and 1, it is treated as the proportion of N's allowed in a read. "
            "(no)")
    
    group = parser.add_argument_group("Output")
    group.add_argument("--quiet", default=False, action='store_true',
        help="Print only error messages. (no)")
    group.add_argument("-o", "--output", metavar="FILE",
        help="Write trimmed reads to FILE. FASTQ or FASTA format is chosen "
            "depending on input. The summary report is sent to standard output. "
            "Use '{name}' in FILE to demultiplex reads into multiple "
            "files. (write to standard output)")
    group.add_argument("--info-file", metavar="FILE",
        help="Write information about each read and its adapter matches into FILE. "
             "See the documentation for the file format. (no)")
    group.add_argument("-r", "--rest-file", metavar="FILE",
        help="When the adapter matches in the middle of a read, write the "
            "rest (after the adapter) into FILE. (no)")
    group.add_argument("--wildcard-file", metavar="FILE",
        help="When the adapter has N bases (wildcards), write adapter bases "
            "matching wildcard positions to FILE. When there are indels in the "
            "alignment, this will often not be accurate. (no)")
    group.add_argument("--too-short-output", metavar="FILE",
        help="Write reads that are too short (according to length specified by "
        "-m) to FILE. (no - too short reads are discarded)")
    group.add_argument("--too-long-output", metavar="FILE",
        help="Write reads that are too long (according to length specified by "
        "-M) to FILE. (no - too long reads are discarded)")
    group.add_argument("--untrimmed-output", default=None, metavar="FILE",
        help="Write reads that do not contain the adapter to FILE. (no - "
             "untrimmed reads are written to default output)")
    group.add_argument("--merged-output", default=None, metavar="FILE",
        help="Write reads that have been merged to this file. By default, merged "
             "reads are written to --too-short-output if specified, and otherwise "
             "are discarded. (no - merged reads are written to default output)")
    group.add_argument("--report-file", default=None, metavar="FILE",
        help="Write report to file rather than stdout/stderr. (no)")
    group.add_argument("--log-level", default=None,
        choices=("DEBUG", "INFO", "WARN", "ERROR"),
        help="Logging level. (ERROR when --quiet else INFO)")
    
    group = parser.add_argument_group("Colorspace options")
    group.add_argument("-c", "--colorspace", action='store_true', default=False,
        help="Enable colorspace mode: Also trim the color that is adjacent to "
             "the found adapter. (no)")
    group.add_argument("-d", "--double-encode", action='store_true', default=False,
        help="Double-encode colors (map 0,1,2,3,4 to A,C,G,T,N). (no)")
    group.add_argument("-t", "--trim-primer", action='store_true', default=False,
        help="Trim primer base and the first color (which is the transition "
            "to the first nucleotide). (no)")
    group.add_argument("--strip-f3", action='store_true', default=False,
        help="Strip the _F3 suffix of read names. (no)")
    group.add_argument("--maq", "--bwa", action='store_true', default=False,
        help="MAQ- and BWA-compatible colorspace output. This enables -c, -d, "
            "-t, --strip-f3 and -y '/1'. (no)")
    group.add_argument("--no-zero-cap", dest='zero_cap', action='store_false',
        help="Do not change negative quality values to zero in colorspace "
            "data. By default, they are since many tools have problems with "
            "negative qualities. (no)")
    group.add_argument("--zero-cap", "-z", action='store_true',
        help="Change negative quality values to zero. This is enabled "
        "by default when -c/--colorspace is also enabled. Use the above option "
        "to disable it. (no)")
    parser.set_defaults(zero_cap=None, action='trim')
    
    group = parser.add_argument_group("Paired-end options", description="The "
        "-A/-G/-B/-U options work like their -a/-b/-g/-u counterparts, but "
        "are applied to the second read in each pair.")
    group.add_argument("-A", dest='adapters2', action='append', default=[],
        metavar='ADAPTER',
        help="3' adapter to be removed from second read in a pair. (no)")
    group.add_argument("-G", dest='front2', action='append', default=[],
        metavar='ADAPTER',
        help="5' adapter to be removed from second read in a pair. (no)")
    group.add_argument("-B", dest='anywhere2', action='append', default=[],
        metavar='ADAPTER',
        help="5'/3 adapter to be removed from second read in a pair. (no)")
    group.add_argument("-U", dest='cut2', action='append', default=[], type=int,
        metavar="LENGTH",
        help="Remove LENGTH bases from second read in a pair (see --cut). (no)")
    group.add_argument("-I", "--cut-min2", action='append', default=[],
        type=int, metavar="LENGTH",
        help="Similar to -U, except that cutting is done AFTER adapter trimming, "
             "and only if a minimum of LENGTH bases was not already removed "
             "(see --cut-min). (no)")
    group.add_argument("-p", "--paired-output", metavar="FILE",
        help="Write second read in a pair to FILE. (no)")
    # Setting the default for pair_filter to None allows us to find out whether
    # the option was used at all.
    group.add_argument("--pair-filter", metavar='(any|both)', default=None,
        choices=("any", "both"),
        help="Which of the reads in a paired-end read have to match the "
            "filtering criterion in order for it to be filtered. (any)")
    group.add_argument("--interleaved", action='store_true', default=False,
        help="Read and write interleaved paired-end reads. (no)")
    group.add_argument("--untrimmed-paired-output", metavar="FILE",
        help="Write second read in a pair to this FILE when no adapter "
            "was found in the first read. Use this option together with "
            "--untrimmed-output when trimming paired-end reads. (no - output "
            "to same file as trimmed reads)")
    group.add_argument("--too-short-paired-output", metavar="FILE", default=None,
        help="Write second read in a pair to this file if pair is too short. "
            "Use together with --too-short-output. (no - too short reads are "
            "discarded)")
    group.add_argument("--too-long-paired-output", metavar="FILE", default=None,
        help="Write second read in a pair to this file if pair is too long. "
            "Use together with --too-long-output. (no - too long reads are "
            "discarded)")
    
    group = parser.add_argument_group("Method-specific options")
    group.add_argument("--bisulfite", default=False, metavar="METHOD",
        help="Set default option values for bisulfite-treated data. The argument specifies the "
             "type of bisulfite library (non-directional, truseq, epignome, or swift) or custom "
             "parameters for trimming: '<read1>[;<read2>]' where trimming parameters for each "
             "read are: '<5' cut>,<3' cut>,<include trimmed>,<only trimmed>' where 'include "
             "trimmed' is 1 or 0 for whether or not the bases already trimmed during/prior to adapter "
             "trimming should be counted towards the total bases to be cut and 'only trimmed' "
             "is 1 or 0 for whether or not only trimmed reads should be further cut. (no)")
    group.add_argument("--mirna", action="store_true", default=False,
        help="Set default option values for miRNA data. (no)")
    
    group = parser.add_argument_group("Parallel (multi-core) options")
    group.add_argument("-T", "--threads", type=int, default=None, metavar="THREADS",
        help="Number of threads to use for read trimming. Set to 0 to use "
             "max available threads. (Do not use multithreading)")
    group.add_argument("--no-writer-process", action="store_true", default=False,
        help="Do not use a writer process; instead, each worker thread writes its "
             "own output to a file with a '.N' suffix. (no)")
    group.add_argument("--preserve-order", action="store_true", default=False,
        help="Preserve order of reads in input files (ignored if "
             "--no-writer-process is set). (no)")
    group.add_argument("--process-timeout", type=int, default=30, metavar="SECONDS",
        help="Number of seconds process should wait before escalating messages "
             "to ERROR level. (30)")
    group.add_argument("--parallel-environment", choices=("local", "cluster"), default="local",
        help="Automatically set performance-tuning parameters based on the environment in which "
             "the program is running. local: assumes fast file I/O and more limited memory; "
             "cluster: assumes slower (i.e. network-based) file I/O and greater memory availablity. "
             "(local)")
    
    group = parser.add_argument_group("Performance-tuning options")
    group.add_argument("--batch-size", type=int, default=5000, metavar="SIZE",
        help="Number of records to process in each batch. (5000)")
    group.add_argument("--read-queue-size", type=int, default=None, metavar="SIZE",
        help="Size of queue for batches of reads to be processed. (THREADS * 100 "
             "in local mode, and THREADS * 1000 in cluster mode)")
    group.add_argument("--result-queue-size", type=int, default=None, metavar="SIZE",
        help="Size of queue for batches of results to be written. (THREADS * 100 "
             "in local mode, and THREADS * 1000 in cluster mode)")
    group.add_argument("--worker-compression", action="store_true", default=False,
        help="Perform data compression in the worker processes rather than the "
             "writer process. (no)")
    
    return parser

def setup_logging(stdout=False, level="INFO"):
    """
    Attach handler to the global logger object
    """
    # Due to backwards compatibility, logging output is sent to standard output
    # instead of standard error if the -o option is used.
    level = getattr(logging, level)
    stream_handler = logging.StreamHandler(sys.stdout if stdout else sys.stderr)
    stream_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s: %(message)s'))
    stream_handler.setLevel(level)
    logging.getLogger().setLevel(level)
    logging.getLogger().addHandler(stream_handler)

MAGNITUDE = dict(
    G=(1E9, "000000000"),
    M=(1E6, "000000"),
    K=(1E3, "000")
)

def validate_options(options, parser):
    if sum(int(opt is not False) for opt in (options.mirna, options.bisulfite)) > 1:
        parser.error("Cannot specify more than one method-specific option")
    
    if options.debug and options.threads is not None:
        parser.error("Cannot use debug mode with multiple threads")
    
    args = options.inputs
    if len(args) == 0:
        parser.error("At least one parameter needed: name of a FASTA or FASTQ file.")
    elif len(args) > 2:
        parser.error("Too many parameters.")
    if args[0].endswith('.qual'):
        parser.error("If a .qual file is given, it must be the second argument.")
    
    # Find out which 'mode' we need to use.
    # Default: single-read trimming (neither -p nor -A/-G/-B/-U/--interleaved given)
    paired = False
    if options.paired_output:
        # Modify first read only, keep second in sync (-p given, but not -A/-G/-B/-U).
        # This exists for backwards compatibility ('legacy mode').
        paired = 'first'
    # Any of these options switch off legacy mode
    if (options.adapters2 or options.front2 or options.anywhere2 or
        options.cut2 or options.cut_min2 or options.interleaved or options.pair_filter or
        options.too_short_paired_output or options.too_long_paired_output):
        # Full paired-end trimming when both -p and -A/-G/-B/-U given
        # Read modifications (such as quality trimming) are applied also to second read.
        paired = 'both'

    if paired and len(args) == 1 and not options.interleaved:
        parser.error("When paired-end trimming is enabled via -A/-G/-B/-U or -p, "
            "two input files are required.")
    if options.interleaved and len(args) != 1:
        parser.error("When reading interleaved files, only one input file may "
            "be given.")
    if not paired:
        if options.untrimmed_paired_output:
            parser.error("Option --untrimmed-paired-output can only be used when "
                "trimming paired-end reads (with option -p).")

    if paired:
        if not options.interleaved:
            if not options.paired_output:
                parser.error("When paired-end trimming is enabled via -A/-G/-B/-U, "
                    "a second output file needs to be specified via -p (--paired-output).")
            if not options.output:
                parser.error("When you use -p or --paired-output, you must also "
                    "use the -o option.")
            if bool(options.untrimmed_output) != bool(options.untrimmed_paired_output):
                parser.error("When trimming paired-end reads, you must use either none "
                    "or both of the --untrimmed-output/--untrimmed-paired-output options.")
            if options.too_short_output and not options.too_short_paired_output:
                parser.error("When using --too-short-output with paired-end "
                    "reads, you also need to use --too-short-paired-output")
            if options.too_long_output and not options.too_long_paired_output:
                parser.error("When using --too-long-output with paired-end "
                    "reads, you also need to use --too-long-paired-output")
    elif len(args) == 2:
        if options.format is not None:
            parser.error("If a pair of .fasta and .qual files is given, the -f/--format "
                "parameter cannot be used.")
    
    if options.merge_overlapping and (
            paired != "both" or
            options.adapters is None or
            options.adapters2 is None or
            options.times != 1):
        parser.error("--merge-overlapping currently only works for paired-end reads with 3' adapters.")
    
    if options.format is not None and options.format.lower() not in ['fasta', 'fastq', 'sra-fastq']:
        parser.error("The input file format must be either 'fasta', 'fastq' or "
            "'sra-fastq' (not '{0}').".format(options.format))
    
    if options.mirna and options.bisulfite is not None:
        parser.error("Can only specify one of --mirna, --bisulfite")
    if options.mirna:
        if options.adapter is None and options.front is None and options.anywhere is None:
            options.adapter = ['TGGAATTCTCGG'] # illumina small RNA adapter
        if options.quality_cutoff is None:
            options.quality_cutoff = "20,20"
        if options.minimum_length is None:
            options.minimum_length = 16
        if options.error_rate is None:
            options.error_rate = 0.12
    elif options.bisulfite:
        if options.quality_cutoff is None:
            options.quality_cutoff = "20,20"
        if options.bisulfite not in ("non-directional", "truseq", "epignome", "swift"):
            def parse_bisulfite_params(r):
                try:
                    parts = [int(p) for p in r.split(",")]
                    assert len(parts) == 4
                    if parts[0] <= 0 and parts[1] <= 0:
                        return None
                    return dict(zip(
                        ("lengths", "count_trimmed", "only_trimmed"),
                        ((parts[0], -1 * parts[1]), (False,True)[parts[2]], (False,True)[parts[3]])
                    ))
                except:
                    parser.error("Invalidate format for bisulfite parameters")
            
            temp = [parse_bisulfite_params(r) for r in options.bisulfite.split(";")]
            if paired == "both" and len(temp) == 1:
                temp = [temp[0], temp[0]]
            elif paired != "both" and len(temp) > 1:
                parser.error("Too many bisulfite parameters for single-end reads")
            options.bisulfite = temp
        elif not paired and options.bisulfite != "non-directional":
            parser.error("Single-end bisulfite libraries must be set to 'non-directional'.")
    
    if options.quality_cutoff is not None:
        cutoffs = options.quality_cutoff.split(',')
        if len(cutoffs) == 1:
            try:
                cutoffs_array = [0, int(cutoffs[0])]
            except ValueError as e:
                parser.error("Quality cutoff value not recognized: {0}".format(e))
        elif len(cutoffs) == 2:
            try:
                cutoffs_array = [int(cutoffs[0]), int(cutoffs[1])]
            except ValueError as e:
                parser.error("Quality cutoff value not recognized: {0}".format(e))
        else:
            parser.error("Expected one value or two values separated by comma for the quality cutoff")
        
        if all(x <= 0 for x in cutoffs_array):
            options.quality_cutoff = None
        else:
            options.quality_cutoff = cutoffs_array
        
    if options.pair_filter is None:
        options.pair_filter = 'any'

    if (options.discard_trimmed or options.discard_untrimmed) and (options.untrimmed_output is not None):
        parser.error("Only one of the --discard-trimmed, --discard-untrimmed "
            "and --untrimmed-output options can be used at the same time.")
    
    if options.output is not None and '{name}' in options.output:
        if options.discard_trimmed:
            parser.error("Do not use --discard-trimmed when demultiplexing.")
        if paired:
            parser.error("Demultiplexing not supported for paired-end files, yet.")
    
    if options.maq:
        options.colorspace = True
        options.double_encode = True
        options.trim_primer = True
        options.suffix = "/1"
    
    if options.strip_f3 or options.maq:
        options.strip_suffix.append('_F3')
    
    if options.zero_cap is None:
        options.zero_cap = options.colorspace
    
    if options.trim_primer and not options.colorspace:
        parser.error("Trimming the primer makes only sense in colorspace.")
    
    if options.double_encode and not options.colorspace:
        parser.error("Double-encoding makes only sense in colorspace.")
    
    if options.anywhere and options.colorspace:
        parser.error("Using --anywhere with colorspace reads is currently not supported (if you "
            "think this may be useful, contact the author).")
    
    if options.error_rate is None:
        options.error_rate = 0.1
    elif not (0 <= options.error_rate <= 1.):
        parser.error("The maximum error rate must be between 0 and 1.")
    
    if options.overlap < 1:
        parser.error("The overlap must be at least 1.")

    if options.cut:
        if len(options.cut) > 2:
            parser.error("You cannot remove bases from more than two ends.")
        if len(options.cut) == 2 and options.cut[0] * options.cut[1] > 0:
            parser.error("You cannot remove bases from the same end twice.")
    
    if options.cut_min:
        if len(options.cut_min) > 2:
            parser.error("You cannot remove bases from more than two ends.")
        if len(options.cut_min) == 2 and options.cut_min[0] * options.cut_min[1] > 0:
            parser.error("You cannot remove bases from the same end twice.")
    
    if paired == 'both' and options.cut2:
        if len(options.cut2) > 2:
            parser.error("You cannot remove bases from more than two ends.")
        if len(options.cut2) == 2 and options.cut2[0] * options.cut2[1] > 0:
            parser.error("You cannot remove bases from the same end twice.")

    if paired == 'both' and options.cut_min2:
        if len(options.cut_min2) > 2:
            parser.error("You cannot remove bases from more than two ends.")
        if len(options.cut_min2) == 2 and options.cut_min2[0] * options.cut_min2[1] > 0:
            parser.error("You cannot remove bases from the same end twice.")
    
    if options.colorspace:
        if options.match_read_wildcards:
            parser.error('IUPAC wildcards not supported in colorspace')
        options.match_adapter_wildcards = False
    
    if options.quiet or options.output is None or options.output == "-":
        options.progress = None
    elif options.progress == "bar":
        try:
            import progressbar
        except:
            parser.error("The python-progressbar library is required for --progress=bar")
    
    # TODO: once we switch to argparse, int_or_str can be passed
    # as the argument type
    def int_or_str(x):
        if x is None or isinstance(x, int):
            return x
        elif isinstance(x, str):
            x = x.upper()
            for a, mag in MAGNITUDE.items():
                x = x.replace(a, mag[1])
            return int(x)
        else:
            raise Exception("Unsupported type {}".format(x))
    
    options.max_reads = int_or_str(options.max_reads)
    
    if options.threads is not None:
        threads = options.threads
        if threads <= 0:
            threads = cpu_count()
        # reserve one thread each for the reader and writer processes
        threads = threads - (1 if options.no_writer_process else 2)
        if threads <= 0:
            logging.getLogger().warn("Using more than the available number of CPUs")
            threads = 1
        options.threads = threads
        
        options.batch_size = int_or_str(options.batch_size)
        if options.process_timeout < 0:
            options.process_timeout = 0
        
        # set default option values based on running environment
        if options.parallel_environment == "local":
            default_queue_size = options.threads * 100
        elif options.parallel_environment == "cluster":
            default_queue_size = options.threads * 1000
            options.worker_compression = True
        
        # set queue sizes if necessary
        if options.read_queue_size is None:
            options.read_queue_size = default_queue_size
        elif options.read_queue_size > 0:
            assert options.read_queue_size >= options.threads
    
        if options.result_queue_size is None:
            options.result_queue_size = default_queue_size
        elif options.result_queue_size > 0:
            assert options.result_queue_size > options.threads
    
    return paired

def create_reader(input_files, options, paired, parser, counter_magnitude="M"):
    input_filename = input_files[0]
    input_paired_filename = None
    quality_filename = None
    if len(input_files) > 1:
        if paired:
            input_paired_filename = input_files[1]
        else:
            quality_filename = input_files[1]
    
    try:
        reader = open_reader(
            input_filename, file2=input_paired_filename,
            qualfile=quality_filename, colorspace=options.colorspace,
            fileformat=options.format, interleaved=options.interleaved)
    except (UnknownFileType, IOError) as e:
        parser.error(e)
    
    qualities = reader.delivers_qualities
    
    # Wrap reader in batch iterator
    batch_size = options.batch_size or 1000
    reader = BatchIterator(reader, batch_size, options.max_reads)
    
    # Wrap iterator in progress bar
    if options.progress == "msg":
        reader = ProgressMessageReader(reader, batch_size)
    elif options.progress == "bar":
        reader = create_progress_reader(reader, options.max_reads, counter_magnitude)
    
    return (reader, qualities, quality_filename is not None)

class BatchIterator(object):
    def __init__(self, reader, size, max_reads=None):
        self.reader = reader
        self.iterable = enumerate(reader, 1)
        self.size = size
        self.max_reads = max_reads
        self.done = False
        self._empty_batch = [None] * size
    
    def __iter__(self):
        return self
    
    def __next__(self):
        if self.done:
            raise StopIteration()
        
        try:
            read_index, record = next(self.iterable)
        except StopIteration:
            self.close()
            raise
        
        batch = copy.copy(self._empty_batch)
        batch[0] = record
        batch_index = 1
        max_size = self.size
        if self.max_reads:
            max_size = min(max_size, self.max_reads - read_index + 1)
        
        while batch_index < max_size:
            try:
                read_index, record = next(self.iterable)
                batch[batch_index] = record
                batch_index += 1
            except StopIteration:
                self.close()
                break
        
        if self.max_reads and read_index >= self.max_reads:
            self.close()
        
        if batch_index == self.size:
            return (batch_index, batch)
        else:
            return (batch_index, batch[0:batch_index])
    
    # py2x alias
    next = __next__
    
    def close(self):
        self.done = True
        self.reader.close()

class ProgressMessageReader(object):
    def __init__(self, iterable, batch_size, interval=1000000):
        self.iterable = iterable
        self.batch_size = batch_size
        self.interval = interval
        self.ctr = 0
    
    def __next__(self):
        value = self.iterable.next()
        if value:
            self.ctr += value[0]
            if self.ctr % self.interval < self.batch_size:
                now = time.time()
                logging.getLogger().info("Read {0} records in {1:.1f} seconds".format(self.ctr, now - self.start))
        return value
    
    next = __next__
    def __iter__(self):
        self.start = time.time()
        return self
    
    def close(self):
        logging.getLogger().info("Read a total of {} records".format(self.ctr))
        self.iterable.close()
        
def create_progress_reader(reader, max_reads=None, counter_magnitude="M"):
    try:
        return create_progressbar_reader(reader, max_reads, counter_magnitude)
    except:
        pass
    
    try:
        return create_tqdm_reader(reader)
    except:
        pass
    
    logging.getLogger().warn("No progress bar library available")
    return reader

def create_progressbar_reader(reader, max_reads=None, counter_magnitude="M"):
    import progressbar
    import progressbar.widgets
    import math

    class ProgressBarReader(progressbar.ProgressBar):
        def __init__(self, iterable, widgets, max_value=None):
            super(ProgressBarReader, self).__init__(
                widgets=widgets, max_value=max_value or progressbar.UnknownLength)
            self._iterable = iterable
            self.done = False
        
        def __next__(self):
            try:
                value = next(self._iterable)
                if self.start_time is None:
                    self.start()
                self.update(self.value + value[0])
                return value
            except StopIteration:
                self.close()
                raise
        
        def close(self):
            if not self.done:
                self.finish()
                self.done = True
            if not self._iterable.done:
                self._iterable.close()
    
    class MagCounter(progressbar.widgets.WidgetBase):
        def __init__(self, magnitude):
            suffix = ""
            if magnitude is None:
                div = 1.0
            else:
                div = float(MAGNITUDE[magnitude][0])
                suffix = magnitude
        
            self._format = lambda val: "{:.1f} {}".format(val / div, suffix)
    
        def __call__(self, progress, data):
            return self._format(data["value"])
        
    if max_reads:
        reader = ProgressBarReader(reader, [
            MagCounter(counter_magnitude), " Reads (", progressbar.Percentage(), ") ",
            progressbar.Timer(), " ", progressbar.Bar(), progressbar.AdaptiveETA()
        ], max_reads)
    else:
        reader = ProgressBarReader(reader, [
            MagCounter(counter_magnitude), " Reads", progressbar.Timer(),
            progressbar.AnimatedMarker()
        ])
    
    return reader

def create_tqdm_reader(reader):
    import tqdm
    return tqdm.tqdm(reader)

def create_filters(options, paired, min_affected):
    filters = Filters(FilterFactory(paired, min_affected))
    
    if options.minimum_length is not None and options.minimum_length > 0:
        filters.add_filter(TooShortReadFilter, options.minimum_length)
    
    if options.maximum_length < sys.maxsize:
        filters.add_filter(TooLongReadFilter, options.maximum_length)
    
    if options.max_n >= 0:
        filters.add_filter(NContentFilter, options.max_n)
    
    if options.discard_trimmed:
        filters.add_filter(TrimmedFilter)
    
    if options.discard_untrimmed or options.untrimmed_output:
        filters.add_filter(UntrimmedFilter)
    
    return filters

def create_modifiers(options, paired, qualities, has_qual_file, parser):
    adapter_parser = AdapterParser(
        colorspace=options.colorspace,
        max_error_rate=options.error_rate,
        min_overlap=options.overlap,
        read_wildcards=options.match_read_wildcards,
        adapter_wildcards=options.match_adapter_wildcards,
        indels=options.indels)

    try:
        adapters1 = adapter_parser.parse_multi(options.adapters, options.anywhere, options.front)
        adapters2 = adapter_parser.parse_multi(options.adapters2, options.anywhere2, options.front2)
    except IOError as e:
        if e.errno == errno.ENOENT:
            parser.error(e)
        raise
    except ValueError as e:
        parser.error(e)
    
    if not adapters1 and not adapters2 and not options.quality_cutoff and \
            options.nextseq_trim is None and \
            options.cut == [] and options.cut2 == [] and \
            options.cut_min == [] and options.cut_min2 == [] and \
            (options.minimum_length is None or options.minimum_length <= 0) and \
            options.maximum_length == sys.maxsize and \
            not has_qual_file and \
            options.max_n == -1 and not options.trim_n:
        parser.error("You need to provide at least one adapter sequence.")
    
    if options.debug:
        for adapter in adapters1 + adapters2:
            adapter.enable_debug()
    
    merger = None
    if options.merge_overlapping:
        merger = MergeOverlapping(
            min_overlap=options.merge_min_overlap,
            error_rate=options.error_rate)
    
    modifiers = Modifiers(paired, merger)
    
    if options.cut or options.cut2:
        modifiers.add_modifier(UnconditionalCutter, read=1, lengths=options.cut)
        modifiers.add_modifier(UnconditionalCutter, read=2, lengths=options.cut2)
    
    if options.nextseq_trim is not None:
        modifiers.add_modifier(NextseqQualityTrimmer, read=1,
                               cutoff=options.nextseq_trim, base=options.quality_base)
    
    if options.quality_cutoff:
        modifiers.add_modifier(QualityTrimmer, cutoff_front=options.quality_cutoff[0],
                               cutoff_back=options.quality_cutoff[1], base=options.quality_base)
    
    if adapters1 or adapters2:
        modifiers.add_modifier(AdapterCutter, read=1, adapters=adapters1, times=options.times, action=options.action)
        modifiers.add_modifier(AdapterCutter, read=2, adapters=adapters2, times=options.times, action=options.action)
    
    if options.bisulfite:
        if isinstance(options.bisulfite, str):
            if options.bisulfite == "non_directional":
                modifiers.add_modifier(NonDirectionalBisulfiteTrimmer)
            elif options.bisulfite == "epignome":
                modifiers.add_modifier(EpiGnomeBisulfiteTrimmer)
            elif options.bisulfite == "truseq":
                # TODO: how to trim?
                pass
            elif options.bisulfite == "swift":
                modifiers.add_modifier(SwiftBisulfiteTrimmer)
        else:
            if options.bisulfite[0]:
                modifiers.add_modifier(MinCutter, read=1, **(options.bisulfite[0]))
            if len(options.bisulfite) > 1 and options.bisulfite[1]:
                modifiers.add_modifier(MinCutter, read=2, **(options.bisulfite[1]))
    
    if options.trim_n:
        modifiers.add_modifier(NEndTrimmer)
    
    if options.cut_min or options.cut_min2:
        modifiers.add_modifier(MinCutter, read=1, lengths=options.cut_min)
        modifiers.add_modifier(MinCutter, read=2, lengths=options.cut_min2)
    
    if options.length_tag:
        modifiers.add_modifier(LengthTagModifier, length_tag=options.length_tag)
    
    if options.strip_suffix:
        modifiers.add_modifier(SuffixRemover, suffixes=options.strip_suffix)
    
    if options.prefix or options.suffix:
        modifiers.add_modifier(PrefixSuffixAdder, prefix=options.prefix, suffix=options.suffix)
    
    if options.double_encode:
        modifiers.add_modifier(DoubleEncoder)
    
    if options.zero_cap and qualities:
        modifiers.add_modifier(ZeroCapper, quality_base=options.quality_base)
    
    if options.trim_primer:
        modifiers.add_modifier(PrimerTrimmer)
    
    return (modifiers, len(adapters1) + len(adapters2))

def create_formatters(options, qualities, default_outfile):
    seq_formatter_args = dict(
        qualities=qualities,
        colorspace=options.colorspace,
        interleaved=options.interleaved
    )
    formatters = Formatters(options.output, seq_formatter_args)
    force_create = []
        
    if (options.merge_overlapping and options.merged_output):
        formatters.add_seq_formatter(MergedReadFilter, options.merged_output)
        
    if (options.minimum_length is not None
            and options.minimum_length > 0
            and options.too_short_output):
        formatters.add_seq_formatter(TooShortReadFilter,
            options.too_short_output, options.too_short_paired_output)

    if options.maximum_length < sys.maxsize and options.too_long_output is not None:
        formatters.add_seq_formatter(TooLongReadFilter,
            options.too_long_output, options.too_long_paired_output)
    
    if not formatters.multiplexed:
        if options.output is not None:
            formatters.add_seq_formatter(NoFilter,
                options.output, options.paired_output)
            if options.output != "-":
                force_create.append(options.output)
                if options.paired_output is not None:
                    force_create.append(options.paired_output)
        elif not (options.discard_trimmed and options.untrimmed_output):
            formatters.add_seq_formatter(NoFilter, default_outfile)
            if default_outfile != "-":
                force_create.append(default_outfile)
    
    if not options.discard_untrimmed:
        if formatters.multiplexed:
            untrimmed = options.untrimmed_output or options.output.format(name='unknown')
            formatters.add_seq_formatter(UntrimmedFilter, untrimmed)
            formatters.add_seq_formatter(NoFilter, untrimmed)
        elif options.untrimmed_output:
            formatters.add_seq_formatter(UntrimmedFilter,
                options.untrimmed_output, options.untrimmed_paired_output)

    if options.rest_file:
        formatters.add_info_formatter(RestFormatter(options.rest_file))
    if options.info_file:
        formatters.add_info_formatter(InfoFormatter(options.info_file))
    if options.wildcard_file:
        formatters.add_info_formatter(WildcardFormatter(options.wildcard_file))
    
    return (formatters, force_create)

def run_serial(reader, modifiers, filters, formatters, writers):
    try:
        n = 0
        total_bp1 = 0
        total_bp2 = 0
        
        for batch_size, batch in reader:
            n += batch_size
            result = defaultdict(lambda: [])
            for record in batch:
                reads, bp = modifiers.modify(record)
                total_bp1 += bp[0]
                total_bp2 += bp[1]
                dest = filters.filter(*reads)
                formatters.format(result, dest, *reads)
            result = dict((path, "".join(strings))
                for path, strings in result.items())
            writers.write_result(result)
        
        return (0, Summary(
            collect_process_statistics(n, total_bp1, total_bp2, modifiers, filters, formatters),
            summarize_adapters(modifiers.get_modifier_pair(AdapterCutter)),
            modifiers.get_trimmer_classes()
        ).finish())
    
    except KeyboardInterrupt as e:
        logging.getLogger().error("Interrupted")
        return (130, None)
    except IOError as e:
        if e.errno == errno.EPIPE:
            return (1, None)
        raise
    except (FormatError, EOFError) as e:
        logging.getLogger().error("Cutadapt error", exc_info=True)
        return(1, None)
    finally:
        reader.close()
        writers.close()

if __name__ == '__main__':
    main()



# coding: utf-8
from __future__ import print_function, division, absolute_import

from atropos.align import (locate, compare_prefixes, compare_suffixes,
	Aligner)
from atropos.adapters import BACK


class TestAligner():
	def test(self):
		reference = 'CTCCAGCTTAGACATATC'
		aligner = Aligner(reference, 0.1, flags=BACK)
		aligner.locate('CC')

	def test_100_percent_error_rate(self):
		reference = 'GCTTAGACATATC'
		aligner = Aligner(reference, 1.0, flags=BACK)
		aligner.locate('CAA')


def test_polya():
	s = 'AAAAAAAAAAAAAAAAA'
	t = 'ACAGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
	result = locate(s, t, 0.0, BACK)
	#start_s, stop_s, start_t, stop_t, matches, cost = result
	assert result == (0, len(s), 4, 4 + len(s), len(s), 0)


# Sequences with IUPAC wildcards
# R=A|G, Y=C|T, S=G|C, W=A|T, K=G|T, M=A|C, B=C|G|T, D=A|G|T, H=A|C|T, V=A|C|G,
# N=A|C|G|T, X={}
WILDCARD_SEQUENCES = [
	'CCCATTGATC',  # original sequence without wildcards
	'CCCRTTRATC',  # R=A|G
	'YCCATYGATC',  # Y=C|T
	'CSSATTSATC',  # S=G|C
	'CCCWWWGATC',  # W=A|T
	'CCCATKKATC',  # K=G|T
	'CCMATTGMTC',  # M=A|C
	'BCCATTBABC',  # B=C|G|T
	'BCCATTBABC',  # B
	'CCCDTTDADC',  # D=A|G|T
	'CHCATHGATC',  # H=A|C|T
	'CVCVTTVATC',  # V=A|C|G
	'CCNATNGATC',  # N=A|C|G|T
	'CCCNTTNATC',  # N
#   'CCCXTTXATC',  # X
]


def test_compare_prefixes():
	assert compare_prefixes('AAXAA', 'AAAAATTTTTTTTT') == (0, 5, 0, 5, 4, 1)
	assert compare_prefixes('AANAA', 'AACAATTTTTTTTT', wildcard_ref=True) == (0, 5, 0, 5, 5, 0)
	assert compare_prefixes('AANAA', 'AACAATTTTTTTTT', wildcard_ref=True) == (0, 5, 0, 5, 5, 0)
	assert compare_prefixes('XAAAAA', 'AAAAATTTTTTTTT') == (0, 6, 0, 6, 4, 2)

	a = WILDCARD_SEQUENCES[0]
	for s in WILDCARD_SEQUENCES:
		r = s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
		result = compare_prefixes(a, r, wildcard_query=True)
		assert result == (0, 10, 0, 10, 10, 0), result

		result = compare_prefixes(r, a, wildcard_ref=True)
		assert result == (0, 10, 0, 10, 10, 0)

	for s in WILDCARD_SEQUENCES:
		for t in WILDCARD_SEQUENCES:
			r = s + 'GCCAGGG'
			result = compare_prefixes(s, r, )
			assert result == (0, 10, 0, 10, 10, 0)

			result = compare_prefixes(r, s, wildcard_ref=True, wildcard_query=True)
			assert result == (0, 10, 0, 10, 10, 0)

	r = WILDCARD_SEQUENCES[0] + 'GCCAGG'
	for wildc_ref in (False, True):
		for wildc_query in (False, True):
			result = compare_prefixes('CCCXTTXATC', r, wildcard_ref=wildc_ref, wildcard_query=wildc_query)
			assert result == (0, 10, 0, 10, 8, 2)


def test_compare_suffixes():
	assert compare_suffixes('AAXAA', 'TTTTTTTAAAAA') == (0, 5, 7, 12, 4, 1)
	assert compare_suffixes('AANAA', 'TTTTTTTAACAA', wildcard_ref=True) == (0, 5, 7, 12, 5, 0)
	assert compare_suffixes('AANAA', 'TTTTTTTAACAA', wildcard_ref=True) == (0, 5, 7, 12, 5, 0)
	assert compare_suffixes('AAAAAX', 'TTTTTTTAAAAA') == (0, 6, 6, 12, 4, 2)


def test_wildcards_in_adapter():
	r = 'CATCTGTCC' + WILDCARD_SEQUENCES[0] + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
	for a in WILDCARD_SEQUENCES:
		result = locate(a, r, 0.0, BACK, wildcard_ref=True)
		assert result == (0, 10, 9, 19, 10, 0), result

	a = 'CCCXTTXATC'
	result = locate(a, r, 0.0, BACK, wildcard_ref=True)
	assert result is None


def test_wildcards_in_read():
	a = WILDCARD_SEQUENCES[0]
	for s in WILDCARD_SEQUENCES:
		r = 'CATCTGTCC' + s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
		result = locate(a, r, 0.0, BACK, wildcard_query=True)
		if 'X' in s:
			assert result is None
		else:
			assert result == (0, 10, 9, 19, 10, 0), result


def test_wildcards_in_both():
	for a in WILDCARD_SEQUENCES:
		for s in WILDCARD_SEQUENCES:
			if 'X' in s or 'X' in a:
				continue
			r = 'CATCTGTCC' + s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
			result = locate(a, r, 0.0, BACK, wildcard_ref=True, wildcard_query=True)
			assert result == (0, 10, 9, 19, 10, 0), result


def test_no_match():
	a = locate('CTGATCTGGCCG', 'AAAAGGG', 0.1, BACK)
	assert a is None, a



# coding: utf-8
from __future__ import print_function, division, absolute_import
from nose.tools import raises, assert_raises

from atropos.seqio import Sequence
from atropos.adapters import (Adapter, Match, ColorspaceAdapter, FRONT, BACK,
	parse_braces, LinkedAdapter)

def test_issue_52():
	adapter = Adapter(
		sequence='GAACTCCAGTCACNNNNN',
		where=BACK,
		max_error_rate=0.12,
		min_overlap=5,
		read_wildcards=False,
		adapter_wildcards=True)
	read = Sequence(name="abc", sequence='CCCCAGAACTACAGTCCCGGC')
	am = Match(astart=0, astop=17, rstart=5, rstop=21, matches=15, errors=2, front=None, adapter=adapter, read=read)
	assert am.wildcards() == 'GGC'
	"""
	The result above should actually be 'CGGC' since the correct
	alignment is this one:

	adapter         GAACTCCAGTCACNNNNN
	mismatches           X     X
	read       CCCCAGAACTACAGTC-CCGGC

	Since we do not keep the alignment, guessing 'GGC' is the best we
	can currently do.
	"""


def test_issue_80():
	# This issue turned out to not be an actual issue with the alignment
	# algorithm. The following alignment is found because it has more matches
	# than the 'obvious' one:
	#
	# TCGTATGCCGTCTTC
	# =========X==XX=
	# TCGTATGCCCTC--C
	#
	# This is correct, albeit a little surprising, since an alignment without
	# indels would have only two errors.

	adapter = Adapter(
		sequence="TCGTATGCCGTCTTC",
		where=BACK,
		max_error_rate=0.2,
		min_overlap=3,
		read_wildcards=False,
		adapter_wildcards=False)
	read = Sequence(name="seq2", sequence="TCGTATGCCCTCC")
	result = adapter.match_to(read)
	assert read.original_length == 13, result
	assert result.errors == 3, result
	assert result.astart == 0, result
	assert result.astop == 15, result


def test_str():
	a = Adapter('ACGT', where=BACK, max_error_rate=0.1)
	str(a)
	str(a.match_to(Sequence(name='seq', sequence='TTACGT')))
	ca = ColorspaceAdapter('0123', where=BACK, max_error_rate=0.1)
	str(ca)


@raises(ValueError)
def test_color():
	ColorspaceAdapter('0123', where=FRONT, max_error_rate=0.1)


def test_parse_braces():
	assert parse_braces('') == ''
	assert parse_braces('A') == 'A'
	assert parse_braces('A{0}') == ''
	assert parse_braces('A{1}') == 'A'
	assert parse_braces('A{2}') == 'AA'
	assert parse_braces('A{2}C') == 'AAC'
	assert parse_braces('ACGTN{3}TGACCC') == 'ACGTNNNTGACCC'
	assert parse_braces('ACGTN{10}TGACCC') == 'ACGTNNNNNNNNNNTGACCC'
	assert parse_braces('ACGTN{3}TGA{4}CCC') == 'ACGTNNNTGAAAACCC'
	assert parse_braces('ACGTN{0}TGA{4}CCC') == 'ACGTTGAAAACCC'


def test_parse_braces_fail():
	for expression in ['{', '}', '{}', '{5', '{1}', 'A{-7}', 'A{', 'A{1', 'N{7', 'AN{7', 'A{4{}',
			'A{4}{3}', 'A{b}', 'A{6X}', 'A{X6}']:
		assert_raises(ValueError, lambda: parse_braces(expression))


def test_linked_adapter():
	linked_adapter = LinkedAdapter('AAAA', 'TTTT')
	sequence = Sequence(name='seq', sequence='AAAACCCCCTTTT')
	match = linked_adapter.match_to(sequence)
	trimmed = linked_adapter.trimmed(match)
	assert trimmed.name == 'seq'
	assert trimmed.sequence == 'CCCCC'



# cython: profile=False, emit_code_comments=False
from cpython.mem cimport PyMem_Malloc, PyMem_Free, PyMem_Realloc

DEF START_WITHIN_SEQ1 = 1
DEF START_WITHIN_SEQ2 = 2
DEF STOP_WITHIN_SEQ1 = 4
DEF STOP_WITHIN_SEQ2 = 8
DEF SEMIGLOBAL = 15

# structure for a DP matrix entry
ctypedef struct _Entry:
    int cost
    int matches  # no. of matches in this alignment
    int origin   # where the alignment originated: negative for positions within seq1, positive for pos. within seq2


ctypedef struct _Match:
    int origin
    int cost
    int matches
    int ref_stop
    int query_stop


def _acgt_table():
    """
    Return a translation table that maps A, C, G, T characters to the lower
    four bits of a byte. Other characters (including possibly IUPAC characters)
    are mapped to zero.

    Lowercase versions are also translated, and U is treated the same as T.
    """
    d = dict(A=1, C=2, G=4, T=8, U=8)
    t = bytearray(b'\0') * 256
    for c, v in d.items():
        t[ord(c)] = v
        t[ord(c.lower())] = v
    return bytes(t)


def _iupac_table():
    """
    Return a translation table for IUPAC characters.

    The table maps ASCII-encoded IUPAC nucleotide characters to bytes in which
    the four least significant bits are used to represent one nucleotide each.

    Whether two characters x and y match can then be checked with the
    expression "x & y != 0".
    """
    A = 1
    C = 2
    G = 4
    T = 8
    d = dict(
        X=0,
        A=A,
        C=C,
        G=G,
        T=T,
        U=T,
        R=A|G,
        Y=C|T,
        S=G|C,
        W=A|T,
        K=G|T,
        M=A|C,
        B=C|G|T,
        D=A|G|T,
        H=A|C|T,
        V=A|C|G,
        N=A|C|G|T
    )
    t = bytearray(b'\0') * 256
    for c, v in d.items():
        t[ord(c)] = v
        t[ord(c.lower())] = v
    return bytes(t)


cdef bytes ACGT_TABLE = _acgt_table()
cdef bytes IUPAC_TABLE = _iupac_table()


class DPMatrix:
    """
    Representation of the dynamic-programming matrix.

    This used only when debugging is enabled in the Aligner class since the
    matrix is normally not stored in full.

    Entries in the matrix may be None, in which case that value was not
    computed.
    """
    def __init__(self, reference, query):
        m = len(reference)
        n = len(query)
        self._rows = [ [None] * (n+1) for _ in range(m + 1) ]
        self.reference = reference
        self.query = query

    def set_entry(self, int i, int j, cost):
        """
        Set an entry in the dynamic programming matrix.
        """
        self._rows[i][j] = cost

    def __str__(self):
        """
        Return a representation of the matrix as a string.
        """
        rows = ['     ' + ' '.join(c.rjust(2) for c in self.query)]
        for c, row in zip(' ' + self.reference, self._rows):
            r = c + ' ' + ' '.join('  ' if v is None else '{0:2d}'.format(v) for v in row)
            rows.append(r)
        return '\n'.join(rows)


cdef class Aligner:
    """
    TODO documentation still uses s1 (reference) and s2 (query).

    Locate one string within another by computing an optimal semiglobal
    alignment between string1 and string2.

    The alignment uses unit costs, which means that mismatches, insertions and deletions are
    counted as one error.

    flags is a bitwise 'or' of the allowed flags.
    To allow skipping of a prefix of string1 at no cost, set the
    START_WITHIN_SEQ1 flag.
    To allow skipping of a prefix of string2 at no cost, set the
    START_WITHIN_SEQ2 flag.
    If both are set, a prefix of string1 or of string1 is skipped,
    never both.
    Similarly, set STOP_WITHIN_SEQ1 and STOP_WITHIN_SEQ2 to
    allow skipping of suffixes of string1 or string2. Again, when both
    flags are set, never suffixes in both strings are skipped.
    If all flags are set, this results in standard semiglobal alignment.

    The skipped parts are described with two intervals (start1, stop1),
    (start2, stop2).

    For example, an optimal semiglobal alignment of SISSI and MISSISSIPPI looks like this:

    ---SISSI---
    MISSISSIPPI

    start1, stop1 = 0, 5
    start2, stop2 = 3, 8
    (with zero errors)

    The aligned parts are string1[start1:stop1] and string2[start2:stop2].

    The error rate is: errors / length where length is (stop1 - start1).

    An optimal alignment fulfills all of these criteria:

    - its error_rate is at most max_error_rate
    - Among those alignments with error_rate <= max_error_rate, the alignment contains
      a maximal number of matches (there is no alignment with more matches).
    - If there are multiple alignments with the same no. of matches, then one that
      has minimal no. of errors is chosen.
    - If there are still multiple candidates, choose the alignment that starts at the
      leftmost position within the read.

    The alignment itself is not returned, only the tuple
    (start1, stop1, start2, stop2, matches, errors), where the first four fields have the
    meaning as described, matches is the number of matches and errors is the number of
    errors in the alignment.

    It is always the case that at least one of start1 and start2 is zero.

    IUPAC wildcard characters can be allowed in the reference and the query
    by setting the appropriate flags.

    If neither flag is set, the full ASCII alphabet is used for comparison.
    If any of the flags is set, all non-IUPAC characters in the sequences
    compare as 'not equal'.
    """
    cdef int m
    cdef _Entry* column  # one column of the DP matrix
    cdef double max_error_rate
    cdef int flags
    cdef int _insertion_cost
    cdef int _deletion_cost
    cdef int _min_overlap
    cdef bint wildcard_ref
    cdef bint wildcard_query
    cdef bint debug
    cdef object _dpmatrix
    cdef bytes _reference  # TODO rename to translated_reference or so
    cdef str str_reference

    def __cinit__(self, str reference, double max_error_rate, int flags=SEMIGLOBAL, bint wildcard_ref=False, bint wildcard_query=False):
        self.max_error_rate = max_error_rate
        self.flags = flags
        self.wildcard_ref = wildcard_ref
        self.wildcard_query = wildcard_query
        self.str_reference = reference
        self.reference = reference
        self._min_overlap = 1
        self.debug = False
        self._dpmatrix = None
        self._insertion_cost = 1
        self._deletion_cost = 1

    property min_overlap:
        def __get__(self):
            return self._min_overlap

        def __set__(self, int value):
            if value < 1:
                raise ValueError('Minimum overlap must be at least 1')
            self._min_overlap = value

    property indel_cost:
        """
        Matches cost 0, mismatches cost 1. Only insertion/deletion costs can be
        changed.
        """
        def __set__(self, value):
            if value < 1:
                raise ValueError('Insertion/deletion cost must be at leat 1')
            self._insertion_cost = value
            self._deletion_cost = value

    property reference:
        def __get__(self):
            return self._reference

        def __set__(self, str reference):
            mem = <_Entry*> PyMem_Realloc(self.column, (len(reference) + 1) * sizeof(_Entry))
            if not mem:
                raise MemoryError()
            self.column = mem
            self._reference = reference.encode('ascii')
            self.m = len(reference)
            if self.wildcard_ref:
                self._reference = self._reference.translate(IUPAC_TABLE)
            elif self.wildcard_query:
                self._reference = self._reference.translate(ACGT_TABLE)
            self.str_reference = reference

    property dpmatrix:
        """
        The dynamic programming matrix as a DPMatrix object. This attribute is
        usually None, unless debugging has been enabled with enable_debug().
        """
        def __get__(self):
            return self._dpmatrix

    def enable_debug(self):
        """
        Store the dynamic programming matrix while running the locate() method
        and make it available in the .dpmatrix attribute.
        """
        self.debug = True

    def locate(self, str query):
        """
        locate(query) -> (refstart, refstop, querystart, querystop, matches, errors)

        Find the query within the reference associated with this aligner. The
        intervals (querystart, querystop) and (refstart, refstop) give the
        location of the match.

        That is, the substrings query[querystart:querystop] and
        self.reference[refstart:refstop] were found to align best to each other,
        with the given number of matches and the given number of errors.

        The alignment itself is not returned.
        """
        cdef char* s1 = self._reference
        cdef bytes query_bytes = query.encode('ascii')
        cdef char* s2 = query_bytes
        cdef int m = self.m
        cdef int n = len(query)
        cdef _Entry* column = self.column
        cdef double max_error_rate = self.max_error_rate
        cdef bint start_in_ref = self.flags & START_WITHIN_SEQ1
        cdef bint start_in_query = self.flags & START_WITHIN_SEQ2
        cdef bint stop_in_ref = self.flags & STOP_WITHIN_SEQ1
        cdef bint stop_in_query = self.flags & STOP_WITHIN_SEQ2

        if self.wildcard_query:
            query_bytes = query_bytes.translate(IUPAC_TABLE)
            s2 = query_bytes
        elif self.wildcard_ref:
            query_bytes = query_bytes.translate(ACGT_TABLE)
            s2 = query_bytes
        cdef bint compare_ascii = not (self.wildcard_query or self.wildcard_ref)
        """
        DP Matrix:
                   query (j)
                 ----------> n
                |
        ref (i) |
                |
                V
               m
        """
        cdef int i, j

        # maximum no. of errors
        cdef int k = <int> (max_error_rate * m)

        # Determine largest and smallest column we need to compute
        cdef int max_n = n
        cdef int min_n = 0
        if not start_in_query:
            # costs can only get worse after column m
            max_n = min(n, m + k)
        if not stop_in_query:
            min_n = max(0, n - m - k)

        # Fill column min_n.
        #
        # Four cases:
        # not startin1, not startin2: c(i,j) = max(i,j); origin(i, j) = 0
        #     startin1, not startin2: c(i,j) = j       ; origin(i, j) = min(0, j - i)
        # not startin1,     startin2: c(i,j) = i       ; origin(i, j) =
        #     startin1,     startin2: c(i,j) = min(i,j)

        # TODO (later)
        # fill out columns only until 'last'
        if not start_in_ref and not start_in_query:
            for i in range(m + 1):
                column[i].matches = 0
                column[i].cost = max(i, min_n) * self._insertion_cost
                column[i].origin = 0
        elif start_in_ref and not start_in_query:
            for i in range(m + 1):
                column[i].matches = 0
                column[i].cost = min_n * self._insertion_cost
                column[i].origin = min(0, min_n - i)
        elif not start_in_ref and start_in_query:
            for i in range(m + 1):
                column[i].matches = 0
                column[i].cost = i * self._insertion_cost
                column[i].origin = max(0, min_n - i)
        else:
            for i in range(m + 1):
                column[i].matches = 0
                column[i].cost = min(i, min_n) * self._insertion_cost
                column[i].origin = min_n - i

        if self.debug:
            self._dpmatrix = DPMatrix(self.str_reference, query)
            for i in range(m + 1):
                self._dpmatrix.set_entry(i, min_n, column[i].cost)
        cdef _Match best
        best.ref_stop = m
        best.query_stop = n
        best.cost = m + n
        best.origin = 0
        best.matches = 0

        # Ukkonen's trick: index of the last cell that is less than k.
        cdef int last = min(m, k + 1)
        if start_in_ref:
            last = m

        cdef int cost_diag
        cdef int cost_deletion
        cdef int cost_insertion
        cdef int origin, cost, matches
        cdef int length
        cdef bint characters_equal
        cdef _Entry tmp_entry

        with nogil:
            # iterate over columns
            for j in range(min_n + 1, max_n + 1):
                # remember first entry
                tmp_entry = column[0]

                # fill in first entry in this column
                if start_in_query:
                    column[0].origin = j
                else:
                    column[0].cost = j * self._insertion_cost
                for i in range(1, last + 1):
                    if compare_ascii:
                        characters_equal = (s1[i-1] == s2[j-1])
                    else:
                        characters_equal = (s1[i-1] & s2[j-1]) != 0
                    if characters_equal:
                        # Characters match: This cannot be an indel.
                        cost = tmp_entry.cost
                        origin = tmp_entry.origin
                        matches = tmp_entry.matches + 1
                    else:
                        # Characters do not match.
                        cost_diag = tmp_entry.cost + 1
                        cost_deletion = column[i].cost + self._deletion_cost
                        cost_insertion = column[i-1].cost + self._insertion_cost

                        if cost_diag <= cost_deletion and cost_diag <= cost_insertion:
                            # MISMATCH
                            cost = cost_diag
                            origin = tmp_entry.origin
                            matches = tmp_entry.matches
                        elif cost_insertion <= cost_deletion:
                            # INSERTION
                            cost = cost_insertion
                            origin = column[i-1].origin
                            matches = column[i-1].matches
                        else:
                            # DELETION
                            cost = cost_deletion
                            origin = column[i].origin
                            matches = column[i].matches

                    # remember current cell for next iteration
                    tmp_entry = column[i]

                    column[i].cost = cost
                    column[i].origin = origin
                    column[i].matches = matches
                if self.debug:
                    with gil:
                        for i in range(last + 1):
                            self._dpmatrix.set_entry(i, j, column[i].cost)
                while last >= 0 and column[last].cost > k:
                    last -= 1
                # last can be -1 here, but will be incremented next.
                # TODO if last is -1, can we stop searching?
                if last < m:
                    last += 1
                elif stop_in_query:
                    # Found a match. If requested, find best match in last row.
                    # length of the aligned part of the reference
                    length = m + min(column[m].origin, 0)
                    cost = column[m].cost
                    matches = column[m].matches
                    if length >= self._min_overlap and cost <= length * max_error_rate and (matches > best.matches or (matches == best.matches and cost < best.cost)):
                        # update
                        best.matches = matches
                        best.cost = cost
                        best.origin = column[m].origin
                        best.ref_stop = m
                        best.query_stop = j
                        if cost == 0 and matches == m:
                            # exact match, stop early
                            break
                # column finished

        if max_n == n:
            first_i = 0 if stop_in_ref else m
            # search in last column # TODO last?
            for i in range(first_i, m+1):
                length = i + min(column[i].origin, 0)
                cost = column[i].cost
                matches = column[i].matches
                if length >= self._min_overlap and cost <= length * max_error_rate and (matches > best.matches or (matches == best.matches and cost < best.cost)):
                    # update best
                    best.matches = matches
                    best.cost = cost
                    best.origin = column[i].origin
                    best.ref_stop = i
                    best.query_stop = n
        if best.cost == m + n:
            # best.cost was initialized with this value.
            # If it is unchanged, no alignment was found that has
            # an error rate within the allowed range.
            return None

        cdef int start1, start2
        if best.origin >= 0:
            start1 = 0
            start2 = best.origin
        else:
            start1 = -best.origin
            start2 = 0

        assert best.ref_stop - start1 > 0  # Do not return empty alignments.
        return (start1, best.ref_stop, start2, best.query_stop, best.matches, best.cost)

    def __dealloc__(self):
        PyMem_Free(self.column)


def locate(str reference, str query, double max_error_rate, int flags=SEMIGLOBAL, bint wildcard_ref=False, bint wildcard_query=False, int min_overlap=1):
    aligner = Aligner(reference, max_error_rate, flags, wildcard_ref, wildcard_query)
    aligner.min_overlap = min_overlap
    return aligner.locate(query)


def compare_prefixes(str ref, str query, bint wildcard_ref=False, bint wildcard_query=False):
    """
    Find out whether one string is the prefix of the other one, allowing
    IUPAC wildcards in ref and/or query if the appropriate flag is set.

    This is used to find an anchored 5' adapter (type 'FRONT') in the 'no indels' mode.
    This is very simple as only the number of errors needs to be counted.

    This function returns a tuple compatible with what Aligner.locate outputs.
    """
    cdef int m = len(ref)
    cdef int n = len(query)
    cdef bytes query_bytes = query.encode('ascii')
    cdef bytes ref_bytes = ref.encode('ascii')
    cdef char* r_ptr
    cdef char* q_ptr
    cdef int length = min(m, n)
    cdef int i, matches = 0
    cdef bint compare_ascii = False

    if wildcard_ref:
        ref_bytes = ref_bytes.translate(IUPAC_TABLE)
    elif wildcard_query:
        ref_bytes = ref_bytes.translate(ACGT_TABLE)
    else:
        compare_ascii = True
    if wildcard_query:
        query_bytes = query_bytes.translate(IUPAC_TABLE)
    elif wildcard_ref:
        query_bytes = query_bytes.translate(ACGT_TABLE)

    if compare_ascii:
        for i in range(length):
            if ref[i] == query[i]:
                matches += 1
    else:
        r_ptr = ref_bytes
        q_ptr = query_bytes
        for i in range(length):
            if (r_ptr[i] & q_ptr[i]) != 0:
                matches += 1

    # length - matches = no. of errors
    return (0, length, 0, length, matches, length - matches)



# coding: utf-8
from __future__ import print_function, division, absolute_import
from multiprocessing import Process, Queue
import time
import io
from atropos.multicore import *

from nose.tools import raises

class TimeoutException(Exception): pass

log_capture_string = None

def setup():
    ### Create the logger
    logger = logging.getLogger('basic_logger')
    logger.setLevel(logging.DEBUG)

    ### Setup the console handler with a StringIO object
    log_capture_string = io.StringIO()
    ch = logging.StreamHandler(log_capture_string)
    ch.setLevel(logging.DEBUG)

    ### Optionally add a formatter
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    ch.setFormatter(formatter)

    ### Add the console handler to the logger
    logger.addHandler(ch)

def assert_log(expected):
    log_contents = log_capture_string.getvalue()
    log_capture_string.close()
    assert log_contents == expected

def test_wait_on():
    class callbacks():
        def __init__(self):
            self.i = 0
            self.j = 0
            
        def condition(self):
            self.i += 1
            return self.i >= 5
            
        def fail_callback(self):
            self.j += 1
    c = callbacks()
    wait_on(c.condition, message="waiting", fail_callback=c.fail_callback)
    assert c.i == 5
    assert c.j == 4
    #assert_log("DEBUG: waiting for {} seconds".format())

@raises(TimeoutException)
def test_timeout():
    def condition(): return False
    wait_on(condition, timeout=2, wait=1, timeout_callback=TimeoutException)

def test_enqueue_dequeue():
    q = Queue(1)
    enqueue(q, 1)
    assert dequeue(q) == 1

@raises(TimeoutException)
def test_enqueue_timeout():
    q = Queue(1)
    q.put(1)
    enqueue(q, 2, timeout=1, block_timeout=2, timeout_callback=TimeoutException)

@raises(TimeoutException)
def test_dequeue_timeout():
    dequeue(Queue(1), timeout=1, block_timeout=2, timeout_callback=TimeoutException)

# TODO: port tests from testparallel here
# Test worker vs writer compression
# Test without writer process



# coding: utf-8
"""
Alignment module.
"""
from atropos._align import Aligner, compare_prefixes, locate

# flags for global alignment

# The interpretation of the first flag is:
# An initial portion of seq1 may be skipped at no cost.
# This is equivalent to saying that in the alignment,
# gaps in the beginning of seq2 are free.
#
# The other flags have an equivalent meaning.
START_WITHIN_SEQ1 = 1
START_WITHIN_SEQ2 = 2
STOP_WITHIN_SEQ1 = 4
STOP_WITHIN_SEQ2 = 8

# Use this to get regular semiglobal alignment
# (all gaps in the beginning or end are free)
SEMIGLOBAL = START_WITHIN_SEQ1 | START_WITHIN_SEQ2 | STOP_WITHIN_SEQ1 | STOP_WITHIN_SEQ2


def compare_suffixes(s1, s2, wildcard_ref=False, wildcard_query=False):
    """
    Find out whether one string is the suffix of the other one, allowing
    mismatches. Used to find an anchored 3' adapter when no indels are allowed.
    """
    s1 = s1[::-1]
    s2 = s2[::-1]
    _, length, _, _, matches, errors = compare_prefixes(s1, s2, wildcard_ref, wildcard_query)
    return (len(s1) - length, len(s1), len(s2) - length, len(s2), matches, errors)



"""
Build atropos.

Cython is run when
* no pre-generated C sources are found,
* or the pre-generated C sources are out of date,
* or when --cython is given on the command line.
"""
import sys
import os.path

from setuptools import setup, Extension
from distutils.version import LooseVersion
from distutils.command.sdist import sdist as _sdist
from distutils.command.build_ext import build_ext as _build_ext

MIN_CYTHON_VERSION = '0.24'

if sys.version_info < (3, 3):
    sys.stdout.write("At least Python 3.3 is required.\n")
    sys.exit(1)


# set __version__
with open(os.path.join(os.path.dirname(__file__), 'atropos', '__init__.py')) as f:
    for line in f:
        if line.startswith('__version__'):
            exec(line)
            break


def out_of_date(extensions):
    """
    Check whether any pyx source is newer than the corresponding generated
    C source or whether any C source is missing.
    """
    for extension in extensions:
        for pyx in extension.sources:
            path, ext = os.path.splitext(pyx)
            if ext not in ('.pyx', '.py'):
                continue
            if extension.language == 'c++':
                csource = path + '.cpp'
            else:
                csource = path + '.c'
            # When comparing modification times, allow five seconds slack:
            # If the installation is being run from pip, modification
            # times are not preserved and therefore depends on the order in
            # which files were unpacked.
            if not os.path.exists(csource) or (
                os.path.getmtime(pyx) > os.path.getmtime(csource) + 5):
                return True
    return False


def no_cythonize(extensions, **_ignore):
    """
    Change file extensions from .pyx to .c or .cpp.

    Copied from Cython documentation
    """
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in ('.pyx', '.py'):
                if extension.language == 'c++':
                    ext = '.cpp'
                else:
                    ext = '.c'
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources


def check_cython_version():
    """Exit if Cython was not found or is too old"""
    try:
        from Cython import __version__ as cyversion
    except ImportError:
        sys.stdout.write(
            "ERROR: Cython is not installed. Install at least Cython version " +
            str(MIN_CYTHON_VERSION) + " to continue.\n")
        sys.exit(1)
    if LooseVersion(cyversion) < LooseVersion(MIN_CYTHON_VERSION):
        sys.stdout.write(
            "ERROR: Your Cython is at version '" + str(cyversion) +
            "', but at least version " + str(MIN_CYTHON_VERSION) + " is required.\n")
        sys.exit(1)


extensions = [
    Extension('atropos._align', sources=['atropos/_align.pyx']),
    Extension('atropos._qualtrim', sources=['atropos/_qualtrim.pyx']),
    Extension('atropos._seqio', sources=['atropos/_seqio.pyx']),
]


class build_ext(_build_ext):
    def run(self):
        # If we encounter a PKG-INFO file, then this is likely a .tar.gz/.zip
        # file retrieved from PyPI that already includes the pre-cythonized
        # extension modules, and then we do not need to run cythonize().
        if os.path.exists('PKG-INFO'):
            no_cythonize(extensions)
        else:
            # Otherwise, this is a 'developer copy' of the code, and then the
            # only sensible thing is to require Cython to be installed.
            check_cython_version()
            from Cython.Build import cythonize
            self.extensions = cythonize(self.extensions)
        _build_ext.run(self)


class sdist(_sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        from Cython.Build import cythonize
        check_cython_version()
        cythonize(extensions)
        _sdist.run(self)

setup(
    name = 'atropos',
    version = __version__,
    author = 'John Didion',
    author_email = 'john.didion@nih.gov',
    url = 'https://atropos.readthedocs.org/',
    description = 'trim adapters from high-throughput sequencing reads',
    license = 'Original Cutadapt code is under MIT license; improvements and additions are in the Public Domain',
    cmdclass = {'sdist': sdist, 'build_ext': build_ext},
    ext_modules = extensions,
    packages = ['atropos', 'atropos.scripts'],
    scripts = ['bin/atropos'],
    extras_require = {
        'progressbar' : ['progressbar2']
    },
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "License :: Public Domain",
        "Natural Language :: English",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ]
)



# coding: utf-8
"""
Quality trimming.
"""
import sys

def quality_trim_index(qualities, cutoff, base=33):
    """
    Find the position at which to trim a low-quality end from a nucleotide sequence.

    Qualities are assumed to be ASCII-encoded as chr(qual + base).

    The algorithm is the same as the one used by BWA within the function
    'bwa_trim_read':
    - Subtract the cutoff value from all qualities.
    - Compute partial sums from all indices to the end of the sequence.
    - Trim sequence at the index at which the sum is minimal.
    """
    s = 0
    max_qual = 0
    max_i = len(qualities)
    for i in reversed(range(max_i)):
        q = ord(qualities[i]) - base
        s += cutoff - q
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            max_i = i
    return max_i


def nextseq_trim_index(sequence, cutoff, base=33):
    """
    Variant of the above quality trimming routine that works on NextSeq data.
    With Illumina NextSeq, bases are encoded with two colors. 'No color' (a
    dark cycle) usually means that a 'G' was sequenced, but that also occurs
    when sequencing falls off the end of the fragment. The read then contains
    a run of high-quality G bases in the end.

    This routine works as the one above, but counts qualities belonging to 'G'
    bases as being equal to cutoff - 1.
    """
    bases = sequence.sequence
    qualities = sequence.qualities
    s = 0
    max_qual = 0
    max_i = len(qualities)
    for i in reversed(xrange(max_i)):
        q = ord(qualities[i]) - base
        if bases[i] == 'G':
            q = cutoff - 1
        s += cutoff - q
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            max_i = i
    return max_i

try:
    from atropos._qualtrim import quality_trim_index, nextseq_trim_index
except:
    pass



# coding: utf-8
"""
Sequence I/O classes: Reading and writing of FASTA and FASTQ files.

TODO

- Sequence.name should be Sequence.description or so (reserve .name for the part
  before the first space)
"""
import re
import sys
import io
from os.path import splitext
from .filters import NoFilter
from .xopen import xopen, open_output

def _shorten(s, n=100):
    """Shorten string s to at most n characters, appending "..." if necessary."""
    if s is None:
        return None
    if len(s) > n:
        s = s[:n-3] + '...'
    return s

class FormatError(Exception):
    """
    Raised when an input file (FASTA or FASTQ) is malformatted.
    """

class UnknownFileType(Exception):
    """
    Raised when open could not autodetect the file type.
    """

## Reading sequences from files ##

class SequenceReader(object):
    """Read possibly compressed files containing sequences"""
    _close_on_exit = False

    def __init__(self, file):
        """
        file is a path or a file-like object. In both cases, the file may
        be compressed (.gz, .bz2, .xz).
        """
        if isinstance(file, str):
            file = xopen(file)
            self._close_on_exit = True
        self._file = file

    def close(self):
        if self._close_on_exit and self._file is not None:
            self._file.close()
            self._file = None

    def __enter__(self):
        if self._file is None:
            raise ValueError("I/O operation on closed SequenceReader")
        return self

    def __exit__(self, *args):
        self.close()

try:
    from ._seqio import Sequence, FastqReader
except ImportError:
    pass

class ColorspaceSequence(Sequence):
    def __init__(self, name, sequence, qualities, primer=None, name2='', original_length=None, match=None, match_info=None, clipped=[0,0,0]):
        # In colorspace, the first character is the last nucleotide of the primer base
        # and the second character encodes the transition from the primer base to the
        # first real base of the read.
        if primer is None:
            self.primer = sequence[0:1]
            sequence = sequence[1:]
        else:
            self.primer = primer
        if qualities is not None and len(sequence) != len(qualities):
            rname = _shorten(name)
            raise FormatError("In read named {0!r}: length of colorspace quality "
                "sequence ({1}) and length of read ({2}) do not match (primer "
                "is: {3!r})".format(rname, len(qualities), len(sequence), self.primer))
        super(ColorspaceSequence, self).__init__(name, sequence, qualities, name2, original_length, match, match_info, clipped)
        if not self.primer in ('A', 'C', 'G', 'T'):
            raise FormatError("Primer base is {0!r} in read {1!r}, but it "
                "should be one of A, C, G, T.".format(
                    self.primer, _shorten(name)))

    def __repr__(self):
        qstr = ''
        if self.qualities is not None:
            qstr = ', qualities={0!r}'.format(_shorten(self.qualities))
        return '<ColorspaceSequence(name={0!r}, primer={1!r}, sequence={2!r}{3})>'.format(
            _shorten(self.name), self.primer, _shorten(self.sequence), qstr)

    def __getitem__(self, key):
        return self.__class__(
            self.name,
            self.sequence[key],
            self.qualities[key] if self.qualities is not None else None,
            self.primer,
            self.name2,
            self.original_length,
            self.match,
            self.match_info,
            self.clipped)

def sra_colorspace_sequence(name, sequence, qualities, name2):
    """Factory for an SRA colorspace sequence (which has one quality value too many)"""
    return ColorspaceSequence(name, sequence, qualities[1:], name2=name2)

class FileWithPrependedLine(object):
    """
    A file-like object that allows to "prepend" a single
    line to an already opened file. That is, further
    reads on the file will return the provided line and
    only then the actual content. This is needed to solve
    the problem of autodetecting input from a stream:
    As soon as the first line has been read, we know
    the file type, but also that line is "gone" and
    unavailable for further processing.
    """
    def __init__(self, file, line):
        """
        file is an already opened file-like object.
        line is a single string (newline will be appended if not included)
        """
        if not line.endswith('\n'):
            line += '\n'
        self.first_line = line
        self._file = file

    def __iter__(self):
        yield self.first_line
        for line in self._file:
            yield line

    def close(self):
        self._file.close()

class FastaReader(SequenceReader):
    """
    Reader for FASTA files.
    """
    def __init__(self, file, keep_linebreaks=False, sequence_class=Sequence):
        """
        file is a path or a file-like object. In both cases, the file may
        be compressed (.gz, .bz2, .xz).

        keep_linebreaks -- whether to keep newline characters in the sequence
        """
        super(FastaReader, self).__init__(file)
        self.sequence_class = sequence_class
        self.delivers_qualities = False
        self._delimiter = '\n' if keep_linebreaks else ''

    def __iter__(self):
        """
        Read next entry from the file (single entry at a time).
        """
        name = None
        seq = []
        for i, line in enumerate(self._file):
            # strip() also removes DOS line breaks
            line = line.strip()
            if not line:
                continue
            if line and line[0] == '>':
                if name is not None:
                    yield self.sequence_class(name, self._delimiter.join(seq), None)
                name = line[1:]
                seq = []
            elif line and line[0] == '#':
                continue
            elif name is not None:
                seq.append(line)
            else:
                raise FormatError("At line {0}: Expected '>' at beginning of "
                    "FASTA record, but got {1!r}.".format(i+1, _shorten(line)))

        if name is not None:
            yield self.sequence_class(name, self._delimiter.join(seq), None)

class ColorspaceFastaReader(FastaReader):
    def __init__(self, file, keep_linebreaks=False):
        super(ColorspaceFastaReader, self).__init__(file, keep_linebreaks,
            sequence_class=ColorspaceSequence)

class ColorspaceFastqReader(FastqReader):
    def __init__(self, file):
        super(ColorspaceFastqReader, self).__init__(file, sequence_class=ColorspaceSequence)

class SRAColorspaceFastqReader(FastqReader):
    def __init__(self, file):
        super(SRAColorspaceFastqReader, self).__init__(file, sequence_class=sra_colorspace_sequence)

class FastaQualReader(object):
    """
    Reader for reads that are stored in .(CS)FASTA and .QUAL files.
    """
    delivers_qualities = True

    def __init__(self, fastafile, qualfile, sequence_class=Sequence):
        """
        fastafile and qualfile are filenames or file-like objects.
        If a filename is used, then .gz files are recognized.

        The objects returned when iteritng over this file are instances of the
        given sequence_class.
        """
        self.fastareader = FastaReader(fastafile)
        self.qualreader = FastaReader(qualfile, keep_linebreaks=True)
        self.sequence_class = sequence_class

    def __iter__(self):
        """
        Yield Sequence objects.
        """
        # conversion dictionary: maps strings to the appropriate ASCII-encoded character
        conv = dict()
        for i in range(-5, 256 - 33):
            conv[str(i)] = chr(i + 33)
        for fastaread, qualread in zip(self.fastareader, self.qualreader):
            if fastaread.name != qualread.name:
                raise FormatError("The read names in the FASTA and QUAL file "
                    "do not match ({0!r} != {1!r})".format(fastaread.name, qualread.name))
            try:
                qualities = ''.join([conv[value] for value in qualread.sequence.split()])
            except KeyError as e:
                raise FormatError("Within read named {0!r}: Found invalid quality "
                    "value {1}".format(fastaread.name, e))
            assert fastaread.name == qualread.name
            yield self.sequence_class(fastaread.name, fastaread.sequence, qualities)

    def close(self):
        self.fastareader.close()
        self.qualreader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

class ColorspaceFastaQualReader(FastaQualReader):
    def __init__(self, fastafile, qualfile):
        super(ColorspaceFastaQualReader, self).__init__(fastafile, qualfile,
            sequence_class=ColorspaceSequence)

def sequence_names_match(r1, r2):
    """
    Check whether the sequences r1 and r2 have identical names, ignoring a
    suffix of '1' or '2'. Some old paired-end reads have names that end in '/1'
    and '/2'. Also, the fastq-dump tool (used for converting SRA files to FASTQ)
    appends a .1 and .2 to paired-end reads if option -I is used.
    """
    name1 = r1.name.split(None, 1)[0]
    name2 = r2.name.split(None, 1)[0]
    if name1[-1:] in '12' and name2[-1:] in '12':
        name1 = name1[:-1]
        name2 = name2[:-1]
    return name1 == name2

class PairedSequenceReader(object):
    """
    Read paired-end reads from two files.

    Wraps two SequenceReader instances, making sure that reads are properly
    paired.
    """
    def __init__(self, file1, file2, colorspace=False, fileformat=None):
        self.reader1 = open_reader(file1, colorspace=colorspace, fileformat=fileformat)
        self.reader2 = open_reader(file2, colorspace=colorspace, fileformat=fileformat)
        self.delivers_qualities = self.reader1.delivers_qualities

    def __iter__(self):
        """
        Iterate over the paired reads. Each item is a pair of Sequence objects.
        """
        # Avoid usage of zip() below since it will consume one item too many.
        it1, it2 = iter(self.reader1), iter(self.reader2)
        while True:
            try:
                r1 = next(it1)
            except StopIteration:
                # End of file 1. Make sure that file 2 is also at end.
                try:
                    next(it2)
                    raise FormatError("Reads are improperly paired. There are more reads in "
                        "file 2 than in file 1.")
                except StopIteration:
                    pass
                break
            try:
                r2 = next(it2)
            except StopIteration:
                raise FormatError("Reads are improperly paired. There are more reads in "
                    "file 1 than in file 2.")
            if not sequence_names_match(r1, r2):
                raise FormatError("Reads are improperly paired. Read name '{0}' "
                    "in file 1 does not match '{1}' in file 2.".format(r1.name, r2.name))
            yield (r1, r2)

    def close(self):
        self.reader1.close()
        self.reader2.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

class InterleavedSequenceReader(object):
    """
    Read paired-end reads from an interleaved FASTQ file.
    """
    def __init__(self, file, colorspace=False, fileformat=None):
        self.reader = open_reader(file, colorspace=colorspace, fileformat=fileformat)
        self.delivers_qualities = self.reader.delivers_qualities

    def __iter__(self):
        # Avoid usage of zip() below since it will consume one item too many.
        it = iter(self.reader)
        for r1 in it:
            try:
                r2 = next(it)
            except StopIteration:
                raise FormatError("Interleaved input file incomplete: Last record has no partner.")
            if not sequence_names_match(r1, r2):
                raise FormatError("Reads are improperly paired. Name {0!r} "
                    "(first) does not match {1!r} (second).".format(r1.name, r2.name))
            yield (r1, r2)

    def close(self):
        self.reader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

def open_reader(file1, file2=None, qualfile=None, colorspace=False, fileformat=None,
                interleaved=False, qualities=None):
    """
    Open sequence files in FASTA or FASTQ format for reading. This is
    a factory that returns an instance of one of the ...Reader
    classes also defined in this module.

    file1, file2, qualfile -- Paths to regular or compressed files or file-like
        objects. Use file1 if data is single-end. If also file2 is provided,
        sequences are paired. If qualfile is given, then file1 must be a FASTA
        file and sequences are single-end. One of file2 and qualfile must always
        be None (no paired-end data is supported when reading qualfiles).

    interleaved -- If True, then file1 contains interleaved paired-end data.
        file2 and qualfile must be None in this case.

    colorspace -- If True, instances of the Colorspace... classes
        are returned.

    fileformat -- If set to None, file format is autodetected from the file name
        extension. Set to 'fasta', 'fastq', or 'sra-fastq' to not auto-detect.
        Colorspace is not auto-detected and must always be requested explicitly.

    qualities -- When mode is 'w' and fileformat is None, this can be set to
        True or False to specify whether the written sequences will have quality
        values. This is is used in two ways:
        * If the output format cannot be determined (unrecognized extension
          etc), no exception is raised, but fasta or fastq format is chosen
          appropriately.
        * When False (no qualities available), an exception is raised when the
          auto-detected output format is FASTQ.

    """
    if interleaved and (file2 is not None or qualfile is not None):
        raise ValueError("When interleaved is set, file2 and qualfile must be None")
    if file2 is not None and qualfile is not None:
        raise ValueError("Setting both file2 and qualfile is not supported")
    
    if file2 is not None:
        return PairedSequenceReader(file1, file2, colorspace, fileformat)

    if interleaved:
        return InterleavedSequenceReader(file1, colorspace, fileformat)
        
    if qualfile is not None:
        if colorspace:
            # read from .(CS)FASTA/.QUAL
            return ColorspaceFastaQualReader(file1, qualfile)
        else:
            return FastaQualReader(file1, qualfile)
    
    fastq_handler = ColorspaceFastqReader if colorspace else FastqReader
    fasta_handler = ColorspaceFastaReader if colorspace else FastaReader

    if fileformat is None and file1 != "-":
        fileformat = guess_format_from_name(file1)
    
    if fileformat is None:
        if file1 == "-":
            file1 = sys.stdin if mode == 'r' else sys.stdout
        for line in file1:
            if line.startswith('#'):
                # Skip comment lines (needed for csfasta)
                continue
            if line.startswith('>'):
                fileformat = 'fasta'
            elif line.startswith('@'):
                fileformat = 'fastq'
            file1 = FileWithPrependedLine(file1, line)
            break
        
    fileformat = fileformat.lower()
    if fileformat == 'fastq' and qualities is False:
        raise ValueError("Output format cannot be FASTQ since no quality values are available.")
    if fileformat == 'fasta':
        return fasta_handler(file1)
    if fileformat == 'fastq':
        return fastq_handler(file1)
    if fileformat == 'sra-fastq' and colorspace:
        return SRAColorspaceFastqReader(file1)
    
    raise UnknownFileType(
        "File format {0!r} is unknown (expected 'sra-fastq' (only for colorspace), "
        "'fasta' or 'fastq').".format(fileformat))

def guess_format_from_name(file, raise_on_failure=False):
    # Detect file format
    name = None
    if isinstance(file, str):
        name = file
    elif hasattr(file, "name"):	 # seems to be an open file-like object
        name = file.name
    
    if name:
        name, ext1, ext2 = _splitext(name)
        ext = ext1.lower()
        if ext in ['.fasta', '.fa', '.fna', '.csfasta', '.csfa']:
            return 'fasta'
        elif ext in ['.fastq', '.fq'] or (ext == '.txt' and name.endswith('_sequence')):
            return 'fastq'
    
    if raise_on_failure:
        raise UnknownFileType("Could not determine whether file {0!r} is FASTA "
            "or FASTQ: file name extension {1!r} not recognized".format(file, ext))

def add_suffix_to_path(path, suffix):
    """
    Add the suffix (str or int) after the file name but
    before the extension.
    """
    name, ext1, ext2 = _splitext(name)
    return "{}{}{}{}".format(name, suffix, ext1, ext2 or "")
    
def _splitext(name):
    ext1 = ext2 = None
    for ext in ('.gz', '.xz', '.bz2'):
        if name.endswith(ext):
            ext2 = ext
            name = name[:-len(ext)]
            break
    name, ext1 = splitext(name)
    return (name, ext1, ext2)

## Converting reads to strings ##

class FastaFormat(object):
    def __init__(self, line_length=None):
        self.text_wrapper = None
        if line_length:
            from textwrap import TextWrapper
            self.text_wrapper = TextWrapper(width=line_length)
    
    def format(self, read):
        return self.format_entry(read.name, read.sequence)
    
    def format_entry(self, name, sequence):
        if self.text_wrapper:
            sequence = self.text_wrapper.fill(sequence)
        return "".join((">", name, "\n", sequence, "\n"))

class ColorspaceFastaFormat(FastaFormat):
    def format(self, read):
        return self.format_entry(read.name, read.primer + read.sequence)

class FastqFormat(object):
    def format(self, read):
        return self.format_entry(read.name, read.sequence, read.qualities, read.name2)
    
    def format_entry(self, name, sequence, qualities, name2=""):
        return "".join((
            '@', name, '\n',
            sequence, '\n+',
            name2, '\n',
            qualities, '\n'
        ))

class ColorspaceFastqFormat(FastqFormat):
    def format(self, read):
        return self.format_entry(read.name, read.primer + read.sequence, read.qualities)

class SingleEndFormatter(object):
    def __init__(self, seq_format, file1):
        self.seq_format = seq_format
        self.file1 = file1
        self.written = 0
        self.read1_bp = 0
        self.read2_bp = 0
    
    def format(self, result, read1, read2=None):
        result[self.file1].append(self.seq_format.format(read1))
        self.written += 1
        self.read1_bp += len(read1)
    
    @property
    def written_bp(self):
        return (self.read1_bp, self.read2_bp)

class InterleavedFormatter(SingleEndFormatter):
    def format(self, result, read1, read2=None):
        result[self.file1].extend((
            self.seq_format.format(read1),
            self.seq_format.format(read2)
        ))
        self.written += 1
        self.read1_bp += len(read1)
        self.read2_bp += len(read2)

class PairedEndFormatter(SingleEndFormatter):
    def __init__(self, seq_format, file1, file2):
        super(PairedEndFormatter, self).__init__(seq_format, file1)
        self.file2 = file2
    
    def format(self, result, read1, read2):
        result[self.file1].append(self.seq_format.format(read1))
        result[self.file2].append(self.seq_format.format(read2))
        self.written += 1
        self.read1_bp += len(read1)
        self.read2_bp += len(read2)

class DelimFormatter(object):
    def __init__(self, path, delim=' '):
        self.path = path
        self.delim = delim
    
    def _format(self, result, fields):
        result[self.path].append("".join((
            self.delim.join(str(f) for f in fields),
            "\n"
        )))

class RestFormatter(DelimFormatter):
    def format(self, result, read):
        if read.match:
            rest = read.match.rest()
            if len(rest) > 0:
                self._format(result, (rest, read.name))

class InfoFormatter(DelimFormatter):
    def __init__(self, path):
        super(InfoFormatter, self).__init__(path, delim='\t')
    
    def format(self, result, read):
        if read.match:
            for m in read.match_info:
                self._format(result, m[0:11])
        else:
            seq = read.sequence
            qualities = read.qualities if read.qualities is not None else ''
            self._format(result, (read.name, -1, seq, qualities))

class WildcardFormatter(DelimFormatter):
    def format(self, result, read):
        if read.match:
            self._format(result, (read.match.wildcards(), read.name))

def create_seq_formatter(file1, file2=None, interleaved=False, **kwargs):
    seq_format = get_format(file1, **kwargs)
    if file2 is not None:
        return PairedEndFormatter(seq_format, file1, file2)
    elif interleaved:
        return InterleavedFormatter(seq_format, file1)
    else:
        return SingleEndFormatter(seq_format, file1)

def get_format(file, fileformat=None, colorspace=False, qualities=None, line_length=None):
    """
    fileformat -- If set to None, file format is autodetected from the file name
        extension. Set to 'fasta', 'fastq', or 'sra-fastq' to not auto-detect.
        Colorspace is not auto-detected and must always be requested explicitly.

    colorspace -- If True, instances of the Colorspace... formats
        are returned.
    
    qualities -- When fileformat is None, this can be set to
        True or False to specify whether the written sequences will have quality
        values. This is is used in two ways:
        * If the output format cannot be determined (unrecognized extension
          etc), no exception is raised, but fasta or fastq format is chosen
          appropriately.
        * When False (no qualities available), an exception is raised when the
          auto-detected output format is FASTQ.
    """
    if fileformat is None:
        fileformat = guess_format_from_name(file, raise_on_failure=qualities is None)
    
    if fileformat is None:
        if qualities is True:
            # Format not recognized, but know we want to write reads with qualities
            fileformat = 'fastq'
        elif qualities is False:
            # Same, but we know that we want to write reads without qualities
            fileformat = 'fasta'
    
    if fileformat is None:
        raise UnknownFileType("Could not determine file type.")
    
    if fileformat == 'fastq' and qualities is False:
        raise ValueError("Output format cannot be FASTQ since no quality values are available.")
    
    fileformat = fileformat.lower()
    if fileformat == 'fasta':
        if colorspace:
            return ColorspaceFastaFormat(line_length)
        else:
            return FastaFormat(line_length)
    elif fileformat == 'fastq':
        if colorspace:
            return ColorspaceFastqFormat()
        else:
            return FastqFormat()
    else:
        raise UnknownFileType("File format {0!r} is unknown (expected 'fasta' or 'fastq').".format(fileformat))

class Formatters(object):
    def __init__(self, output, seq_formatter_args):
        self.output = output
        self.multiplexed = output is not None and '{name}' in output
        self.seq_formatter_args = seq_formatter_args
        self.seq_formatters = {}
        self.mux_formatters = {}
        self.info_formatters = []
        self.discarded = 0
    
    def add_seq_formatter(self, filter_type, file1, file2=None):
        self.seq_formatters[filter_type] = create_seq_formatter(
            file1, file2, **self.seq_formatter_args)
    
    def add_info_formatter(self, formatter):
        self.info_formatters.append(formatter)
    
    def get_mux_formatter(self, name):
        assert self.multiplexed
        if name not in self.mux_formatters:
            path = self.output.format(name=name)
            self.mux_formatters[name] = create_seq_formatter(path, **self.seq_formatter_args)
        return self.mux_formatters[name]
    
    def get_seq_formatters(self):
        return (
            set(f for f in self.seq_formatters.values() if f.written > 0) |
            set(f for f in self.mux_formatters.values() if f.written > 0)
        )
    
    def summary(self):
        seq_formatters = self.get_seq_formatters()
        written = sum(f.written for f in seq_formatters)
        read_bp = [
            sum(f.read1_bp for f in seq_formatters),
            sum(f.read2_bp for f in seq_formatters)
        ]
        return (written, read_bp)
    
    def format(self, result, dest, read1, read2=None):
        if self.multiplexed and (dest == NoFilter) and read1.match:
            name = read1.match.adapter.name
            formatter = self.get_mux_formatter(name)
            formatter.format(result, read1, read2)
        elif dest in self.seq_formatters:
            self.seq_formatters[dest].format(result, read1, read2)
        else:
            self.discarded += 1
        
        for f in self.info_formatters:
            f.format(result, read1)
            if read2:
                f.format(result, read2)

## Writing sequences to files ##

class Writers(object):
    def __init__(self, force_create):
        self.writers = {}
        self.force_create = force_create
        self.suffix = None
    
    def write_result(self, result, compressed=False):
        """
        Write results to file.
        
        result :: dict with keys being file descriptors
        and values being data (either bytes or strings).
        Strings are expected to already have appropriate
        line-endings.
        compressed :: whether data has already been compressed.
        """
        for file_desc, data in result.items():
            self.write(file_desc, data, compressed)
    
    def write(self, file_desc, data, compressed=False):
        """
        Write data to file.
        """
        if compressed:
            path, mode = file_desc
        else:
            path = file_desc
        if path not in self.writers:
            if self.suffix:
                real_path = add_suffix_to_path(path, self.suffix)
            else:
                real_path = path
            # TODO: test whether O_NONBLOCK allows non-blocking write to NFS
            if compressed:
                self.writers[path] = open_output(real_path, mode)
            else:
                self.writers[path] = xopen(real_path, "w")
        self.writers[path].write(data)
    
    def close(self):
        for path in self.force_create:
            if path not in self.writers:
                with open_output(path, "w"):
                    pass
        for writer in self.writers.values():
            writer.close()



# Copyright (c) Twisted Matrix Laboratories.
#
# Copied from Twisted (http://twistedmatrix.com/), see
# http://twistedmatrix.com/trac/browser/trunk/LICENSE for the license.
#
# This makes sure that users don't have to set up their environment
# specially in order to run these programs from bin/.

# This helper is shared by many different actual scripts.  It is not intended to
# be packaged or installed, it is only a developer convenience.  By the time
# the package is actually installed somewhere, the environment should already be set
# up properly without the help of this tool.

import sys, os

path = os.path.abspath(sys.argv[0])
while os.path.dirname(path) != path:
    if os.path.exists(os.path.join(path, 'cutadapt', '__init__.py')):
        sys.path.insert(0, path)
        break
    path = os.path.dirname(path)



# coding: utf-8
"""
Routines for printing a report.
"""
import sys
from collections import OrderedDict, defaultdict
from contextlib import contextmanager
import textwrap
from .adapters import BACK, FRONT, PREFIX, SUFFIX, ANYWHERE, LINKED
from .modifiers import *
from .filters import *

# TODO: Would be good to have a more generic way for filters
# and modifiers to report summary information

def check_equal_merger(dest, src):
    assert dest == src
    return dest

def nested_dict_merger(d_dest1, d_src1):
    assert isinstance(d_src1, dict)
    for k1, d_src2 in d_src1.items():
        if k1 in d_dest1:
            d_dest2 = d_dest1[k1]
            for k2, v_src in d_src2.items():
                if k2 in d_dest2:
                    d_dest2[k2] += v_src
                else:
                    d_dest2[k2] = v_src
        else:
            d_dest1[k1] = d_src2
    return d_dest1

MERGERS = dict(
    check_equal=check_equal_merger,
    nested_dict=nested_dict_merger
)

class MergingDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        super(MergingDict, self).__init__(*args, **kwargs)
        self.mergers = {}
    
    def set_with_merger(self, key, value, merger):
        self[key] = value
        self.mergers[key] = merger
    
    def merge(self, src):
        for k, v_src in src.items():
            if k in self and self[k] is not None:
                if v_src is None:
                    continue
                v_self = self[k]
                if k in self.mergers:
                    merger = MERGERS[self.mergers[k]]
                    self[k] = merger(v_self, v_src)
                # default behavior: lists have two integers, which are summed;
                # dicts have integer values, which are summed; strings must be
                # identical; otherwise must be numeric and are summed
                elif isinstance(v_self, dict):
                    assert isinstance(v_src, dict)
                    for kk,vv in v_src.items():
                        if kk in v_self:
                            v_self[kk] += vv
                        else:
                            v_self[kk] = vv
                elif isinstance(v_self, list):
                    assert isinstance(v_src, list)
                    self[k] = [v_src[0]+v_self[0], v_src[1]+v_self[1]]
                elif isinstance(v_self, str):
                    assert v_self == v_src
                else:
                    self[k] = v_src + v_self
            else:
                self[k] = v_src
                if isinstance(src, MergingDict) and k in src.mergers:
                    self.mergers[k] = src.mergers[k]

def collect_process_statistics(N, total_bp1, total_bp2, modifiers, filters, formatters):
    """
    Collect statistics from filters, and modifiers
    """
    stats = defaultdict(lambda: [0,0])
    
    written, written_bp = formatters.summary()
    assert written is not None
    
    stats.update(dict(
        N=N,
        total_bp1=total_bp1,
        total_bp2=total_bp2,
        total_bp=total_bp1 + total_bp2,
        written=written,
        written_bp=written_bp,
        total_written_bp=sum(written_bp)
    ))
    
    stats["too_short"] = None
    if TooShortReadFilter in filters:
        stats["too_short"] = filters[TooShortReadFilter].filtered
    
    stats["too_long"] = None
    if TooLongReadFilter in filters:
        stats["too_long"] = filters[TooLongReadFilter].filtered
    
    stats["too_many_n"] = None
    if NContentFilter in filters:
        stats["too_many_n"] = filters[NContentFilter].filtered
    
    stats["with_adapters"] = [0, 0]
    for i in (0, 1):
        if AdapterCutter in modifiers[i]:
            stats["with_adapters"][i] += modifiers[i][AdapterCutter].with_adapters
    
    for modifier in modifiers[0].values():
        if isinstance(modifier, Trimmer):
            key = "{}_bp".format(type(modifier).__name__)
            stats[key][0] = modifier.trimmed_bases
            
    for modifier in modifiers[1].values():
        if isinstance(modifier, Trimmer):
            key = "{}_bp".format(type(modifier).__name__)
            stats[key][1] = modifier.trimmed_bases
            
    for modifier_class in modifiers.get_trimmer_classes():
        name = modifier_class.__name__
        stats[name] = sum(stats["{}_bp".format(name)])
    
    return dict(stats)

def summarize_adapters(mods):
    summary = [{}, {}]
    if mods[0] is not None:
        summary[0] = collect_adapter_statistics(mods[0].adapters)
    if mods[1] is not None:
        summary[1] = collect_adapter_statistics(mods[1].adapters)
    return summary

def collect_adapter_statistics(adapters):
    """
    TODO: push this method into Adapter?
    """
    stats = OrderedDict()
    for adapter in adapters:
        name = adapter.name
        total_front = sum(adapter.lengths_front.values())
        total_back = sum(adapter.lengths_back.values())
        
        stats[name] = MergingDict(
            name=name,
            total_front=total_front,
            total_back=total_back,
            total=total_front + total_back
        )
        
        where = adapter.where
        assert (where in (ANYWHERE, LINKED) or
            (where in (BACK, SUFFIX) and total_front == 0) or
            (where in (FRONT, PREFIX) and total_back == 0)
        )
        stats[name].set_with_merger("where", where, "check_equal")
        
        def handle_nested_dict(key, value):
            d = {}
            for k,v in value.items():
                d[k] = dict(v)
            stats[name].set_with_merger(key, d, "nested_dict")
        
        if where == LINKED:
            stats[name]["front_sequence"] = adapter.front_adapter.sequence
            stats[name]["back_sequence"] = adapter.back_adapter.sequence
            stats[name].set_with_merger("front_max_error_rate",
                adapter.front_adapter.max_error_rate, "check_equal")
            stats[name].set_with_merger("back_max_error_rate",
                adapter.back_adapter.max_error_rate, "check_equal")
            stats[name]["front_lengths_front"] = dict(adapter.front_adapter.lengths_front)
            stats[name]["front_lengths_back"] = dict(adapter.front_adapter.lengths_back)
            stats[name]["back_lengths_front"] = dict(adapter.back_adapter.lengths_front)
            stats[name]["back_lengths_back"] = dict(adapter.back_adapter.lengths_back)
            # have to clone these nested dicts and set them
            # up with a custom merge function
            handle_nested_dict("front_errors_front", adapter.front_adapter.errors_front)
            handle_nested_dict("front_errors_back", adapter.front_adapter.errors_back)
            handle_nested_dict("back_errors_front", adapter.back_adapter.errors_front)
            handle_nested_dict("back_errors_back", adapter.back_adapter.errors_back)
        else:
            stats[name]["sequence"] = adapter.sequence
            stats[name].set_with_merger("max_error_rate",
                adapter.max_error_rate, "check_equal")
            if where in (ANYWHERE, FRONT, PREFIX):
                stats[name]["lengths_front"] = dict(adapter.lengths_front)
                handle_nested_dict("errors_front", adapter.errors_front)
            if where in (ANYWHERE, BACK, SUFFIX):
                stats[name]["lengths_back"] = dict(adapter.lengths_back)
                handle_nested_dict("errors_back", adapter.errors_back)
            if where in (BACK, SUFFIX):
                stats[name]["adjacent_bases"] = dict(adapter.adjacent_bases)
    
    return stats

class Summary(object):
    def __init__(self, process_stats=MergingDict(),
                 adapter_stats=[OrderedDict(),OrderedDict()],
                 trimmer_classes=[]):
        self.process_stats = process_stats
        self.adapter_stats = adapter_stats
        self.trimmer_classes = trimmer_classes
    
    def add_process_stats(self, stats):
        self.process_stats.merge(stats)
    
    def add_adapter_stats(self, stats):
        """
        stats = [ {name : adapter_stats}, {name : adapter_stats} ]
        """
        for i in (0,1):
            for name in stats[i].keys():
                if name in self.adapter_stats[i]:
                    self.adapter_stats[i][name].merge(stats[i][name])
                else:
                    self.adapter_stats[i][name] = stats[i][name]
    
    def finish(self):
        stats = self.process_stats.copy()
        
        stats["written_fraction"] = 0
        stats["too_short_fraction"] = 0
        stats["too_long_fraction"] = 0
        stats["too_many_n_fraction"] = 0
        stats["with_adapters_fraction"] = [0, 0]
        stats["total_written_bp_fraction"] = 0.0
        for modifier_class in self.trimmer_classes:
            name = modifier_class.__name__
            stats["{}_fraction".format(name)] = 0.0
            
        N = stats["N"]
        if N > 0:
            stats["written_fraction"] = stats["written"] / N if stats["written"] else 0
            stats["too_short_fraction"] = stats["too_short"] / N if stats["too_short"] else 0
            stats["too_long_fraction"] = stats["too_long"] / N if stats["too_long"] else 0
            stats["too_many_n_fraction"] = stats["too_many_n"] / N if stats["too_many_n"] else 0
            stats["with_adapters_fraction"] = [ (v / N) for v in stats["with_adapters"] ]
        
        if stats["total_bp"] > 0:
            stats["total_written_bp_fraction"] = (stats["total_written_bp"] / stats["total_bp"]) if stats["total_written_bp"] else 0
            for modifier_class in self.trimmer_classes:
                name = modifier_class.__name__
                if stats[name]:
                    stats["{}_fraction".format(name)] = (stats[name] / stats["total_bp"])
        
        stats["adapters"] = [
            self.adapter_stats[0].values(),
            self.adapter_stats[1].values()
        ]
        
        return stats

def print_report(paired, options, wallclock_time, cpu_time, stats, trimmer_classes):
    outfile = options.report_file
    close = False
    
    if outfile is not None:
        outfile = open(outfile, "w")
        close = True
    elif not options.quiet:
        outfile = sys.stderr if options.output is None else sys.stdout
    else:
        return
    
    stats["paired"] = paired
    stats["wallclock_time"] = max(wallclock_time, 0.01)
    stats["cpu_time"] = max(cpu_time, 0.01)
    stats['pairs_or_reads'] = "Pairs" if paired else "Reads"
    
    try:
        # TODO: rewrite generate_report so that this call
        # receives a string result and writes it to outfile
        # outfile.write(generate_report(paired, time, stats))
        generate_report(stats, trimmer_classes, outfile)

    finally:
        if close and outfile is not None:
            outfile.close()
    
        # def get_values(self):
        #	values = var(self)
        #	return values

def generate_report(stats, trimmer_classes, outfile):
    def _print(*args, **kwargs): print(*args, file=outfile, **kwargs)
    
    ADAPTER_TYPES = {
        BACK: "regular 3'",
        FRONT: "regular 5'",
        PREFIX: "anchored 5'",
        SUFFIX: "anchored 3'",
        ANYWHERE: "variable 5'/3'",
        LINKED: "linked",
    }
    
    def print_error_ranges(adapter_length, error_rate):
        _print("No. of allowed errors:")
        prev = 0
        for errors in range(1, int(error_rate * adapter_length) + 1):
            r = int(errors / error_rate)
            _print("{0}-{1} bp: {2};".format(prev, r - 1, errors - 1), end=' ')
            prev = r
        if prev == adapter_length:
            _print("{0} bp: {1}".format(adapter_length, int(error_rate * adapter_length)))
        else:
            _print("{0}-{1} bp: {2}".format(prev, adapter_length, int(error_rate * adapter_length)))
        _print()

    def print_histogram(d, adapter_length, n, error_rate, errors):
        """
        Print a histogram. Also, print the no. of reads expected to be
        trimmed by chance (assuming a uniform distribution of nucleotides in the reads).
        d -- a dictionary mapping lengths of trimmed sequences to their respective frequency
        adapter_length -- adapter length
        n -- total no. of reads.
        """
        h = []
        for length in sorted(d):
            # when length surpasses adapter_length, the
            # probability does not increase anymore
            estimated = n * 0.25 ** min(length, adapter_length)
            h.append( (length, d[length], estimated) )
        
        def errs_to_str(l, e):
            if e in errors[l]:
                return str(errors[l][e])
            return "0"
        
        _print("length", "count", "expect", "max.err", "error counts", sep="\t")
        for length, count, estimate in h:
            max_errors = max(errors[length].keys())
            errs = ' '.join(errs_to_str(length, e) for e in range(max_errors+1))
            _print(length, count, "{0:.1F}".format(estimate), int(error_rate*min(length, adapter_length)), errs, sep="\t")
        _print()

    def print_adjacent_bases(bases, sequence):
        """
        Print a summary of the bases preceding removed adapter sequences.
        Print a warning if one of the bases is overrepresented and there are
        at least 20 preceding bases available.

        Return whether a warning was printed.
        """
        total = sum(bases.values())
        if total == 0:
            return False
        _print('Bases preceding removed adapters:')
        warnbase = None
        for base in ['A', 'C', 'G', 'T', '']:
            b = base if base != '' else 'none/other'
            fraction = 1.0 * bases[base] / total
            _print('  {0}: {1:.1%}'.format(b, fraction))
            if fraction > 0.8 and base != '':
                warnbase = b
        if total >= 20 and warnbase is not None:
            _print('WARNING:')
            _print('	The adapter is preceded by "{0}" extremely often.'.format(warnbase))
            _print('	The provided adapter sequence may be incomplete.')
            _print('	To fix the problem, add "{0}" to the beginning of the adapter sequence.'.format(warnbase))
            _print()
            return True
        _print()
        return False
    
    #----------------
    
    """Print report to standard output."""
    if stats["N"] == 0:
        _print("No reads processed! Either your input file is empty or you used the wrong -f/--format parameter.")
        return
    _print("Wallclock time: {0:.2F} s ({1:.0F} us/read; {2:.2F} M reads/minute).".format(
        stats["wallclock_time"], 1E6 * stats["wallclock_time"] / stats["N"], stats["N"] / stats["wallclock_time"] * 60 / 1E6))
    _print("CPU time (main process): {0:.2F} s".format(stats["cpu_time"]))
    
    report = "\n=== Summary ===\n\n"
    if stats["paired"]:
        report += textwrap.dedent("""\
        Total read pairs processed:		 {N:13,d}
          Read 1 with adapter:			 {with_adapters[0]:13,d} ({with_adapters_fraction[0]:.1%})
          Read 2 with adapter:			 {with_adapters[1]:13,d} ({with_adapters_fraction[1]:.1%})
        """)
    else:
        report += textwrap.dedent("""\
        Total reads processed:			 {N:13,d}
        Reads with adapters:			 {with_adapters[0]:13,d} ({with_adapters_fraction[0]:.1%})
        """)
    if stats["too_short"] is not None:
        report += "{pairs_or_reads} that were too short:		   {too_short:13,d} ({too_short_fraction:.1%})\n"
    if stats["too_long"] is not None:
        report += "{pairs_or_reads} that were too long:			   {too_long:13,d} ({too_long_fraction:.1%})\n"
    if stats["too_many_n"] is not None:
        report += "{pairs_or_reads} with too many N:			   {too_many_n:13,d} ({too_many_n_fraction:.1%})\n"

    report += textwrap.dedent("""\
    {pairs_or_reads} written (passing filters):			{written:13,d} ({written_fraction:.1%})

    Total basepairs processed:               {total_bp:13,d} bp
    """)
    if stats["paired"]:
        report += "	 Read 1: {total_bp1:13,d} bp\n"
        report += "	 Read 2: {total_bp2:13,d} bp\n"
    
    for modifier_class in trimmer_classes:
        name = modifier_class.__name__
        if stats[name] > 0:
            spacing = ' ' * (40 - len(modifier_class.display_str))
            report += "{0}:{1}{{{2}:13,d}} bp ({{{2}_fraction:.1%}})\n".format(modifier_class.display_str, spacing, name)
            if stats["paired"]:
                report += "	 Read 1: {{{}_bp[0]:13,d}} bp\n".format(name)
                report += "	 Read 2: {{{}_bp[1]:13,d}} bp\n".format(name)
    
    report += "Total written (filtered):                {total_written_bp:13,d} bp ({total_written_bp_fraction:.1%})\n"
    if stats["paired"]:
        report += "	 Read 1: {written_bp[0]:13,d} bp\n"
        report += "	 Read 2: {written_bp[1]:13,d} bp\n"
    
    try:
        report = report.format(**stats)
    except ValueError:
        # Python 2.6 does not support the comma format specifier (PEP 378)
        report = report.replace(",d}", "d}").format(**stats)
    _print(report)

    warning = False
    for which_in_pair in (0, 1):
        for adapter in stats["adapters"][which_in_pair]:
            if stats["paired"]:
                extra = 'First read: ' if which_in_pair == 0 else 'Second read: '
            else:
                extra = ''

            _print("=" * 3, extra + "Adapter", adapter["name"], "=" * 3)
            _print()
            if adapter["where"] == LINKED:
                _print("Sequence: {0}...{1}; Type: linked; Length: {2}+{3}; Trimmed: {4} times; Half matches: {5}".
                    format(
                        adapter["front_sequence"],
                        adapter["back_sequence"],
                        len(adapter["front_sequence"]),
                        len(adapter["back_sequence"]),
                        adapter["total_front"], adapter["total_back"]
                    ))
            else:
                _print("Sequence: {0}; Type: {1}; Length: {2}; Trimmed: {3} times.".
                    format(adapter["sequence"], ADAPTER_TYPES[adapter["where"]],
                        len(adapter["sequence"]), adapter["total"]))
    
            if adapter["total"] == 0:
                _print()
                continue
            if adapter["where"] == ANYWHERE:
                _print(adapter["total_front"], "times, it overlapped the 5' end of a read")
                _print(adapter["total_back"], "times, it overlapped the 3' end or was within the read")
                _print()
                print_error_ranges(len(adapter["sequence"]), adapter["max_error_rate"])
                _print("Overview of removed sequences (5')")
                print_histogram(adapter["lengths_front"], len(adapter["sequence"]), stats["N"],
                    adapter["max_error_rate"], adapter["errors_front"])
                _print()
                _print("Overview of removed sequences (3' or within)")
                print_histogram(adapter["lengths_back"], len(adapter["sequence"]), stats["N"],
                    adapter["max_error_rate"], adapter["errors_back"])
            elif adapter["where"] == LINKED:
                _print()
                print_error_ranges(len(adapter["front_sequence"]), adapter["front_max_error_rate"])
                print_error_ranges(len(adapter["back_sequence"]), adapter["back_max_error_rate"])
                _print("Overview of removed sequences at 5' end")
                print_histogram(adapter["front_lengths_front"], len(adapter["front_sequence"]),
                    stats["N"], adapter["front_max_error_rate"], adapter["front_errors_front"])
                _print()
                _print("Overview of removed sequences at 3' end")
                print_histogram(adapter["back_lengths_back"], len(adapter["back_sequence"]),
                    stats["N"], adapter["back_max_error_rate"], adapter["back_errors_back"])
            elif adapter["where"] in (FRONT, PREFIX):
                _print()
                print_error_ranges(len(adapter["sequence"]), adapter["max_error_rate"])
                _print("Overview of removed sequences")
                print_histogram(adapter["lengths_front"], len(adapter["sequence"]),
                    stats["N"], adapter["max_error_rate"], adapter["errors_front"])
            else:
                assert adapter["where"] in (BACK, SUFFIX)
                _print()
                print_error_ranges(len(adapter["sequence"]), adapter["max_error_rate"])
                warning = warning or print_adjacent_bases(adapter["adjacent_bases"], adapter["sequence"])
                _print("Overview of removed sequences")
                print_histogram(adapter["lengths_back"], len(adapter["sequence"]),
                    stats["N"], adapter["max_error_rate"], adapter["errors_back"])
    
    if warning:
        _print('WARNING:')
        _print('	One or more of your adapter sequences may be incomplete.')
        _print('	Please see the detailed output above.')
